//####################################################################
//##     Copyright  Meta Integration Technology, Inc. 1997-2002    ##
//##                      All Rights Reserved.                      ##
//####################################################################
//## Meta Integration(tm) Model Bridge (MIMB), Works (MIW),         ##
//## Repository (MIR), Repository SDK (MIRSDK), and Database SDK    ##
//## (MIDBSDK) are trademarks of Meta Integration Technology, Inc.  ##
//## Other product and company names (or logos) mentioned herein    ##
//## may be the trademarks of their respective owners.              ##
//####################################################################
//##                 http://www.metaintegration.com                 ##
//####################################################################



#ifndef MIRBlob_h
#define MIRBlob_h 1

#include <string.h>

#include "MIRStdDefs.h"


/*
 * This file provides an easy, safe and efficient class to manipulate 
 * binary large objects (blob). There are three classes:
 *
 *  - MIRBlob: This is the class you will use to manipulate blobs.
 *
 *  - MIRTBlob: This is another blob class which is used to 
 *    manipulate transient blobs. It's more efficient in some 
 *    contexts but it is also more dangerous. You can safely ignore 
 *    this class when writing code.
 *
 *  - MIRGBlob: This is the parent class for both MIRBlob and 
 *    MIRTBlob. Its sole purpose is to share code between MIRBlob
 *    and MIRTBlob.
 *
 *
 *                          *** WARNING ***
 * MIRBlob and MIRTBlob behave very differently. Changing the 
 * type of a variable from MIRBlob to MIRTBlob is very likely to 
 * break your program. For all practical purposes though you can use 
 * a MIRBlob, wherever you see a MIRTBlob, i.e. any function that 
 * takes a MIRTBlob as a parameter will also accept a MIRBlob 
 * and the result of a function that returns a MIRTBlob can be put 
 * in a MIRBlob.
 *
 * MIRBlob is designed so that most operations can be performed 
 * naturally. Here are a few examples:
 *
 *  MIRBlob a,b;
 *  a="xxx";        // a == "xxx"
 *  b=a;            // b == "xxx"
 *  a+="-"+b;       // a == "xxx-xxx"
 *
 *
 * Notes:
 *  - when a "char*" parameter is used in the MIRBlob interface and no
 *    size is given, the "char*" must be a null terminated string.
 *  - When a null terminated string is affected to a MIRBlob, the null
 *    terminating character is not copied.
 *  - When comparing a MIRBlob to a null terminated string, the null
 *    terminating character is not compared.
 *
 *  - It's better never to explicitly construct a MIRBlob from a 'char*'. 
 *    This is unnecessary, makes the code less readable and, in some 
 *    cases, may lower the performance. Just write:
 *          extern void process(MIRBlob& blob);
 *          ...
 *          process("See.. no constructor");
 *
 *  - There is one place where you must explicitly cast a MIRBlob to 
 *    'char*': when it is a parameter  
 *    anywhere in the variable part of a function that takes a variable 
 *    number of parameters.
 *
 */

class MIR_STD_API MIRBlob;
class MIR_STD_API MIRTBlob;

/*
 * MIRGBlob class. This is an abstract class used as a parent for both
 * MIRBlob and MIRTBlob.
 */
class MIR_STD_API MIRGBlob
{

      friend class MIRBlob;
      friend class MIRTBlob;

  public:

    //## Comparison

      bool operator==(const MIRGBlob& b) const;
      bool operator!=(const MIRGBlob& b) const;
      bool operator<(const MIRGBlob& b) const;
      bool operator>(const MIRGBlob& b) const;
      bool operator<=(const MIRGBlob& b) const;
      bool operator>=(const MIRGBlob& b) const;
      //    These operators compare two MIRGBlobs and return 
      //    true or false as appropriate.

      bool operator==(const char* cs) const;
      bool operator!=(const char* cs) const;
      bool operator<(const char* cs) const;
      bool operator>(const char* cs) const;  
      bool operator<=(const char* cs) const;
      bool operator>=(const char* cs) const;
      //    These operators compare MIRGBlob to a char* and return 
      //    true or false as appropriate.
      //    The char* buffer must be null terminated. The null character is not compared.

      friend bool MIR_STD_API operator==(const char* cs, const MIRGBlob& b);
      friend bool MIR_STD_API operator!=(const char* cs, const MIRGBlob& b);
      friend bool MIR_STD_API operator<(const char* cs, const MIRGBlob& b);
      friend bool MIR_STD_API operator>(const char* cs, const MIRGBlob& b);
      friend bool MIR_STD_API operator<=(const char* cs, const MIRGBlob& b);
      friend bool MIR_STD_API operator>=(const char* cs, const MIRGBlob& b);
      //    These operators compare a char* to MIRGBlob and return 
      //    true or false as appropriate.
      //    The char* buffer must be null terminated. The null character is not compared.

      int compare(const MIRGBlob& b) const;
      int compare(const char* cs) const;
      int compare(const char* cs, unsigned size) const;
      //    Compares this MIRGBlob to the buffer given as a parameter
      //    and returns:
      //        <0 if   this <  s/cs
      //         0 if   this == s/cs
      //        >0 if   this >  s/cs
      //    The char* buffer must be null terminated if no size is given.

      unsigned int hashCode() const;
      //    Computes a hash code for this blob.


    //## Concatenation

      MIRGBlob& add(const MIRGBlob* first, ...);
      MIRGBlob& add(const char* first, ...);
      //    Concatenates the list of strings given in parameter to the 
      //    end of this buffer and returns a reference to this buffer.
      //    The string list must be terminated by a NULL.
      //    The char* buffers must be null terminated.

      MIRGBlob& add(const MIRGBlob* const * strings);
      MIRGBlob& add(const char* const * strings);
      //    Concatenates the strings in the array to the end of this 
      //    buffer and returns a reference to this buffer.
      //    The string array must be terminated by a NULL.
      //    The char* buffers must be null terminated.

      
    //## General methods


      unsigned setSize(unsigned new_size);
      //    Resizes the buffer so that it can hold a blob of 
      //    length up to 'new_size' characters. If new_size is
      //    less than the current size then the buffer 
      //    is truncated to the specified size. If new_size is
      //    greater than the current size then the last part of the 
      //    buffer (from the current size to the end of the buffer)
      //    is unitialized.
      //    The returned value is the capacity of the buffer after
      //    this call.

      unsigned length() const;
      unsigned getSize() const;
      //    Returns the size allocated for the character buffer.

      void printOn(int level=0) const;
      //    Prints debugging information for this MIRGBlob
      //    object to the standard error stream.


    //## Conversion

      operator char*() const;
      //    Return a pointer to the buffer.

      bool isEmpty() const;
      //    Returns true if the blob is empty, i.e. if its length is 0, 
      //    and false otherwise.

      void empty();
      //    Empties the blob, i.e. sets its length to 0.


    //## Character manipulation

      void setAt(unsigned i, const char c);
      //    Sets the character at the 0-based position i if i < size.
      //    Does nothing otherwise.

      char getAt(unsigned i) const;
      //    Returns the character at the 0-based position i if i < size.
      //    Returns '\0' otherwise.


    //## Character search

      int find(const char c) const;
      int reverseFind(const char c) const;
      //    Respectively find the first and last character c
      //    in the buffer and return its 0-based position.
      //    Returns -1 if not found.

      int find(const char* cs) const;
      //    Finds the first occurence of the substring cs in this
      //    buffer and returns the 0-based position of its first
      //    character.
      //    Returns -1 if not found.

      int find(const MIRGBlob& blob) const;
      //    Finds the first occurence of the blob in this
      //    buffer and returns the 0-based position of its first
      //    character.
      //    Returns -1 if not found.

    //## Substring extraction

      MIRTBlob mid(unsigned int first, unsigned int count) const;
      //    Returns a buffer containing the count characters starting 
      //    from the 0-based position first.
      //    If first >= getSize() or count == 0 mid returns an empty buffer.

      MIRTBlob mid(unsigned int first) const;
      //    Returns a buffer containing the characters starting from 
      //    the 0-based position first to the end of the buffer.
      //    If first >= getSize() mid returns an empty buffer.

      MIRTBlob left(unsigned int count) const;
      //    Returns a buffer containing the count characters starting from 
      //    the beginning of the buffer.
      //    If count == 0 left returns an empty buffer.

      MIRTBlob right(unsigned int count) const;
      //    Returns a buffer containing the count characters starting from 
      //    the end of the buffer.
      //    If count == 0 right returns an empty buffer.


    //## Other transformations

      MIRGBlob& reverse();
      //    Reverses the order of the buffer characters.


    //## Unsafe operations

      /*
       * The following functions are provided for the rare cases where they
       * are needed to achieve optimal performance.
       * Extreme care must be taken when using them.
       */

      void wrap(const MIRGBlob& b);
      //    Initialises this blob so that it refers to the same buffer
      //    as s. You must not change s as this would make this buffer 
      //    inconsistent. Any subsequent change to this blob will cause 
      //    the allocation of a new buffer to store the modified blob.

      void wrap(const char* cs, unsigned size);
      //    Initialises this blob so that it uses cs as its buffer.
      //    You must not change cs as this would make this blob 
      //    inconsistent. Any subsequent change to this blob will cause the 
      //    allocation of a new buffer to store the modified blob. You are 
      //    responsible for freeing cs when appropriate.

      bool unwrap();
      //    If you called wrap before, unwrap will allocate a new buffer to 
      //    store the buffer and return true. This blob will thereafter be 
      //    self-contained as opposed to being a "wrapper blob".
      //    If this blob was not a "wrapper blob" unwrap does nothing and 
      //    returns false.


    //## Miscellaneous friends

      friend MIRTBlob MIR_STD_API operator+(const MIRBlob& b1, const MIRBlob& b2);
      friend MIRTBlob MIR_STD_API operator+(const MIRBlob& b1, const MIRTBlob& b2);
      friend MIRTBlob MIR_STD_API operator+(const MIRBlob& b, const char* cs);
      friend MIRTBlob MIR_STD_API operator+(const char* cs, const MIRBlob& b);
      friend MIRTBlob MIR_STD_API operator+(const MIRTBlob& b1, const MIRBlob& b2);
      friend MIRTBlob MIR_STD_API operator+(const MIRTBlob& b1, const MIRTBlob& b2);
      friend MIRTBlob MIR_STD_API operator+(const MIRTBlob& b, const char* cs);
      friend MIRTBlob MIR_STD_API operator+(const char* cs, const MIRTBlob& b);
      //    These operators have been declared in MIRBlob or MIRTBlob but have to
      //    be redeclared here because they need to be MIRGBlob friends.

  protected:

      char* p;               // character buffer
      unsigned size;         // amount of storage allocated
      bool isWrap;           // a flag to say the blob is a "wrapper blob"
};



/*
 * MIRBlob class.
 */
class MIR_STD_API MIRBlob: public MIRGBlob
{

      friend class MIRTBlob;

  public:

    //## Construction

      MIRBlob();
      MIRBlob(unsigned size);
      //    Creates a blob of the given capacity.
      //    The buffer is unitialized.

      MIRBlob(const char* src);
      MIRBlob(const char* src, unsigned size);
      MIRBlob(const MIRBlob& src);
      MIRBlob(const MIRBlob& src, unsigned size);
      //    Initialises this blob with the contents of src. This implies 
      //    allocating a new buffer and copying the contents of src in this 
      //    new buffer.

      MIRBlob(char c, unsigned size);
      //    Initialises a blob containing size times the character c.

      MIRBlob(const MIRTBlob& src);
      //    Initialises this blob with the contents of src. This implies
      //    taking over the src buffer and making src a wrapper blob for 
      //    this buffer.
      //    See MIRTBlob for more information.

      void add(void *,int);

      static void* operator new(size_t size);
      static void operator delete(void* ptr);
#ifdef WIN32
      static void* operator new(size_t size, const char* file, int line);
      static void* operator new(size_t size, int block_type, const char* file, int line);
#if _MSC_VER >= 1200
      static void operator delete(void* ptr, const char* file, int line);
      static void operator delete(void* ptr, int block_type, const char* file, int line);
#endif
#endif
      //    Used to allocate/deallocate MIRBlob objects (and subclasses).

      void operator=(const MIRBlob& right);
      void operator=(const char* right);
      //    Copies the contents of right to this blob. This implies 
      //    allocating/reallocating a new buffer and copying the contents of 
      //    right in this new buffer.

      void operator=(const MIRTBlob& right);
      //    Assigns right to this blob. This implies taking over the right 
      //    buffer and making right a wrapper blob for this buffer.
      //    See MIRTBlob for more information.

    //## Destruction

      ~MIRBlob();
      //    Releases the resources used by this object.


    //## Concatenation

      const MIRBlob& operator+=(const MIRBlob& right);
      const MIRBlob& operator+=(const MIRTBlob& right);
      const MIRBlob& operator+=(const char* right);
      //    Concatenates the buffer given as a parameter to this 
      //    MIRBlob object and returns a reference to this object.
      //    Note:
      //     - b1+=b2 is much more efficient than b1=b1+b2

      friend MIRTBlob MIR_STD_API operator+(const MIRBlob& b1, const MIRBlob& b2);
      friend MIRTBlob MIR_STD_API operator+(const MIRBlob& b1, const MIRTBlob& b2);
      friend MIRTBlob MIR_STD_API operator+(const MIRBlob& b, const char* cs);
      friend MIRTBlob MIR_STD_API operator+(const char* cs, const MIRBlob& b);
      //    Concatenates the buffer given as a parameter to this 
      //    MIRBlob object and returns a new blob containing the result.


    //## Miscellaneous friends

      friend MIRTBlob MIR_STD_API operator+(const MIRTBlob& b1, const MIRBlob& b2);
      //    This operator is declared in MIRTBlob but has to be redeclared 
      //    here because it needs to be a MIRBlob friend.
};

/*
 * MIRTBlob class.
 *
 * This class is used as an optimised way to store transient blob results. 
 * For instance 'operator+' declares a local variable to store the result of 
 * the concatenation and returns this variable. When returning this variable 
 * the copy constructor is automatically called.
 * Using a MIRBlob for this local variable would cause the copy constructor
 * MIRBlob::MIRBlob(MIRBlob) to be called which means reallocating 
 * a new buffer and copying the local variable's buffer. Right after that 
 * the local variable's destructor is called anyway.
 * Using a MIRTBlob for the local variable MIRBlob::MIRBlob(MIRTBlob) 
 * will be called instead and this copy constructor will not duplicate the 
 * MIRTBlob buffer because it knows that MIRTBlobs are used only once.
 * Instead it takes over the MIRTBlob buffer, makes the MIRTBlob a wrapper
 * around this buffer. This also means that when the MIRTBlob destructor is 
 * called it's almost a no-operation. All in all there is no reallocation or
 * copying involved.
 *
 * All the operations involving a MIRTBlob assume that the MIRTBlob buffer
 * can be modified at will which leads to a very different behavior from a 
 * MIRBlob.
 */
class MIR_STD_API MIRTBlob: public MIRGBlob
{

      friend class MIRBlob;
      friend class MIRGBlob;

  public:

    //## Construction

      MIRTBlob();
      MIRTBlob(unsigned size);
      //    Creates an unitialized transient blob.

      MIRTBlob(const MIRBlob& src);
      MIRTBlob(const MIRTBlob& src);
      MIRTBlob(char* src);
      MIRTBlob(char* src, unsigned size);
      //    Initialises this transient blob so that it refers to the 
      //    same buffer as src. Changing src may make this blob inconsistent.
      //    Any subsequent change to this blob will cause the allocation of
      //    a new buffer to store the modified blob.

      void operator=(const MIRBlob& right);
      void operator=(const MIRTBlob& right);
      void operator=(const char* right);
      //    Initialises this transient blob so that it refers to the 
      //    same buffer as right. Changing right may make this blob
      //    inconsistent. Any subsequent change to this blob will cause
      //    the allocation of a new buffer to store the modified blob.


    //## Destruction

      ~MIRTBlob();
      //    Releases the resources used by this object.


    //## Concatenation

      MIRTBlob& operator+=(const MIRGBlob& right);
      MIRTBlob& operator+=(const char* right);
      //    Concatenates the buffer given as a parameter to this 
      //    MIRTBlob object and returns a reference to this object.
      //    Note:
      //     - b1+=b2 is much more efficient than b1=b1+b2

      friend MIRTBlob MIR_STD_API operator+(const MIRTBlob& b1, const MIRBlob& b2);
      friend MIRTBlob MIR_STD_API operator+(const MIRTBlob& b1, const MIRTBlob& b2);
      friend MIRTBlob MIR_STD_API operator+(const MIRTBlob& b, const char* cs);
      //    Appends the second buffer given in parameter to the MIRTBlob 
      //    object and returns this MIRTBlob parameter.

      friend MIRTBlob MIR_STD_API operator+(const char* cs, const MIRTBlob& b);
      //    Concatenates the MIRTBlob parameter to the end of cs. The result 
      //    is stored in the MIRTBlob object which is then returned.


    //## Miscellaneous friends

      friend MIRTBlob MIR_STD_API operator+(const MIRBlob& b1, const MIRBlob& b2);
      friend MIRTBlob MIR_STD_API operator+(const MIRBlob& b1, const MIRTBlob& b2);
      friend MIRTBlob MIR_STD_API operator+(const MIRBlob& b, const char* cs);
      friend MIRTBlob MIR_STD_API operator+(const char* cs, const MIRBlob& b);
      //    These operators have been declared in MIRBlob but have to
      //    be redeclared here because they need to be MIRTBlob friends.
};


extern MIR_STD_CAPI unsigned int blobHashCode(const char* b, int len);
//    Computes the hash code of the specified blob.


/*
 * MIRGBlob Inline methods
 */

//## MIRGBlob Comparison

inline bool MIRGBlob::operator==(const MIRGBlob& b) const
{
    return (size==b.size) && (size==0 || memcmp(p,b.p,size)==0);
}

inline bool MIRGBlob::operator!=(const MIRGBlob& b) const
{
    return !(*this==b);
}

inline bool MIRGBlob::operator<(const MIRGBlob& b) const
{
    int res = memcmp(p,b.p,size<b.size?size:b.size);
    return res == 0 ? size < b.size : res < 0;
}

inline bool MIRGBlob::operator>(const MIRGBlob& b) const
{
    int res = memcmp(p,b.p,size<b.size?size:b.size);
    return res == 0 ? size > b.size : res > 0;
}

inline bool MIRGBlob::operator<=(const MIRGBlob& b) const
{
    int res = memcmp(p,b.p,size<b.size?size:b.size);
    return res == 0 ? size <= b.size : res < 0;
}

inline bool MIRGBlob::operator>=(const MIRGBlob& b) const
{
    int res = memcmp(p,b.p,size<b.size?size:b.size);
    return res == 0 ? size >= b.size : res > 0;
}

inline bool MIRGBlob::operator==(const char* cs) const
{
    return (size==strlen(cs)) && (size==0 || memcmp(p,cs,size)==0);
}

inline bool MIRGBlob::operator!=(const char* cs) const
{
    return !(*this==cs);
}

inline bool MIRGBlob::operator<(const char* cs) const
{
    unsigned cslen = strlen(cs);
    int res = memcmp(p,cs,size<cslen?size:cslen);
    return res == 0 ? size < cslen : res < 0;
}

inline bool MIRGBlob::operator>(const char* cs) const
{
    unsigned cslen = strlen(cs);
    int res = memcmp(p,cs,size<cslen?size:cslen);
    return res == 0 ? size > cslen : res > 0;
}

inline bool MIRGBlob::operator<=(const char* cs) const
{
    unsigned cslen = strlen(cs);
    int res = memcmp(p,cs,size<cslen?size:cslen);
    return res == 0 ? size <= cslen : res < 0;
}

inline bool MIRGBlob::operator>=(const char* cs) const
{
    unsigned cslen = strlen(cs);
    int res = memcmp(p,cs,size<cslen?size:cslen);
    return res == 0 ? size >= cslen : res > 0;
}

inline bool operator==(const char* cs, const MIRGBlob& b)
{
    return b==cs;
}

inline bool operator!=(const char* cs, const MIRGBlob& b)
{
    return b!=cs;
}

inline bool operator<(const char* cs, const MIRGBlob& b)
{
    return b>=cs;
}

inline bool operator>(const char* cs, const MIRGBlob& b)
{
    return b<=cs;
}

inline bool operator<=(const char* cs, const MIRGBlob& b)
{
    return b>cs;
}

inline bool operator>=(const char* cs, const MIRGBlob& b)
{
    return b<cs;
}

inline int MIRGBlob::compare(const MIRGBlob& b) const
{
    int res = memcmp(p,b.p,size<b.size?size:b.size);
    return res == 0 ? size - b.size : res;
}

inline int MIRGBlob::compare(const char* cs) const
{
    unsigned cslen = strlen(cs);
    int res = memcmp(p,cs,size<cslen?size:cslen);
    return res == 0 ? size - cslen : res;
}

inline int MIRGBlob::compare(const char* cs, unsigned cslen) const
{
    int res = memcmp(p,cs,size<cslen?size:cslen);
    return res == 0 ? size - cslen : res;
}


//## MIRGBlob General methods

inline unsigned MIRGBlob::length() const
{
    return size;
}

inline unsigned MIRGBlob::getSize() const
{
    return size;
}


//## MIRGBlob Conversion

inline MIRGBlob::operator char*() const
{
    return p;
}


//## MIRGBlob Character manipulation

inline void MIRGBlob::setAt(unsigned i, const char c)
{
    if (i<size)
        p[i]=c; 
}

inline char MIRGBlob::getAt(unsigned i) const
{
    return i>=size?'\0':p[i];
}



/*
 * MIRBlob Inline methods
 */

//## MIRBlob Construction

inline MIRBlob::MIRBlob()
{
    p=NULL;
    size=0;
    isWrap=true;
}

#ifdef WIN32
inline void* MIRBlob::operator new(size_t size, const char* file, int line)
{
    return operator new(size,_NORMAL_BLOCK,file,line);
}

#if _MSC_VER >= 1200
inline void MIRBlob::operator delete(void* ptr, int block_type, const char* file, int line)
{
    operator delete(ptr);
}

inline void MIRBlob::operator delete(void* ptr, const char* file, int line)
{
    operator delete(ptr);
}
#endif
#endif





/*
 * MIRTBlob Inline methods
 */

//## MIRTBlob Construction

inline MIRTBlob::MIRTBlob()
{
    p=NULL;
    size=0;
    isWrap=true;
}

inline MIRTBlob::MIRTBlob(const MIRBlob& b)
{
    p=b.p;
    size=b.size;
    isWrap=true;
}

inline MIRTBlob::MIRTBlob(char* cs)
{
    p=cs;
    size=(cs==NULL)?0:strlen(cs);
    isWrap=true;
}

inline MIRTBlob::MIRTBlob(char* cs, unsigned cs_size)
{
    p=cs;
    size=(cs==NULL)?0:cs_size;
    isWrap=true;
}

inline bool MIRGBlob::isEmpty() const
{
    return (size==0);
}

#endif /* end of file */
