//####################################################################
//##     Copyright  Meta Integration Technology, Inc. 1997-2002    ##
//##                      All Rights Reserved.                      ##
//####################################################################
//## Meta Integration(tm) Model Bridge (MIMB), Works (MIW),         ##
//## Repository (MIR), Repository SDK (MIRSDK), and Database SDK    ##
//## (MIDBSDK) are trademarks of Meta Integration Technology, Inc.  ##
//## Other product and company names (or logos) mentioned herein    ##
//## may be the trademarks of their respective owners.              ##
//####################################################################
//##                 http://www.metaintegration.com                 ##
//####################################################################


#ifndef MIRContainer_h
#define MIRContainer_h 1

#include <stdlib.h>

#include "MIRStdDefs.h"


typedef void MIRContainerItem;
typedef void MIRContainerKey;

/*
 * All containers must know their size.
 */
class MIR_STD_API _MIRContainer
{
  public:

    //## Construction

      _MIRContainer();
      //    Creates an empty container.

    //## Destruction

      virtual ~_MIRContainer();


    //## General functions

      int getSize() const;
      //    Returns the number of elements contained in this container.
      //    It is a requirement that this operation be efficient so it
      //    is inlined.

  protected:

      int size;
      //    For performance reasons the number of elements in this 
      //    container is stored (not computed).
};

/*
 * Provides an abstract class defining the common operations 
 * implemented by object pointer containers.
 * To create a container containing a specific type of elements
 * all you need to to is to derive an actual container and redefine 
 * the compare2Object and compare2Key functions.
 *
 * Notes:
 * - Each object put in the container is supposed to have a key.
 *   By default this is the pointer value itself.
 * - Each type of container has an associated iterator type allowing
 *   efficient manipulation of its contents in a container independant 
 *   way.
 * - Containers may or may not preserve the order in which the 
 *   elements have been added. 
 */
 class MIR_STD_API MIRContainer: public virtual _MIRContainer
{
  public:
      class Iterator;

    //## Construction

      MIRContainer();
      //    Creates an empty container.


      //    Actual container implementations should define a constructor
      //    and an operation to set the contents of this container with
	  //    the contents of a container of the same type.

      //MIRContainer(const MIRContainer &src);
      //    Creates a new container initialising it with the contents of 
      //    the given source container. Note that only the object 
      //    pointers are copied, the objects themselves are not duplicated.

      //virtual MIRContainer& operator=(const MIRContainer &right);
      //    Replaces the contents of the current container with that of the
      //    given container. Note that only the object pointers are copied,
      //    the objects themselves are not duplicated.


    //## Destruction

      virtual ~MIRContainer();
      //    Releases the resources used by this container. Note that this 
      //    operation does not release the memory used by the objects 
      //    that have been put in the container.


    //## Query methods

      //    Actual container implementations should define an 
      //    operation to compare the contents of two containers of 
      //    the same type.
      //virtual int operator==(const MIRContainer &right) const;
      //virtual int operator!=(const MIRContainer &right) const;
      //    Tests for (in)equality of the contents of two containers
      //    by comparing the values of the pointers they contain.


    //## General functions

      virtual MIRBoolean empty() = 0;
      //    Removes all the elements that were in the container.
      //    Returns true if successful and false otherwise.


    //## Creating and deleting an iterator for this container

      virtual Iterator* newIterator(MIRBoolean first=true) const = 0;
      //    Allocates an iterator operating on this container. Note 
      //    that each iterator is usually specific to the container type
      //    that alocated it.
      //    If first is true then the iterator is initialised so that it 
      //    points to the first element in the container. Otherwise it 
      //    designates the last element of the container.
      //    Returns a pointer to the new iterator if successful and
      //    NULL otherwise.

      virtual void deleteIterator(Iterator* iterator) const = 0;
      //    Releases an iterator. You must call this method to free any
      //    iterator allocated using the newIterator method. As iterators 
      //    specific to one kind of container be sure to call newIterator
      //    and deleteIterator on the same kind of iterator.


    //## Iterator based methods for adding and removing elements

      virtual MIRBoolean add(const MIRContainerItem* ptr, MIRBoolean unique=true, Iterator* iterator=NULL) = 0;
      //    Adds a new object to the container. If unique is true then
      //    the insertion will fail is an object with the same key
      //    is already in the container.
      //    If the insertion succeeds and iterator is not NULL then it is
      //    set to the position where the object has been inserted and true
      //    is returned. Otherwise, false is returned and iterator is left
      //    unchanged.

      virtual MIRBoolean contains(const MIRContainerItem* ptr, Iterator* iterator=NULL) const = 0;
      //    Returns true if the container contains the object
      //    pointed to by "ptr" and sets the iterator to the position
      //    of this object. Returns false and leaves the iterator 
      //    unchanged otherwise.

      virtual MIRContainerItem* findFirst(const MIRContainerKey* key, Iterator* iterator=NULL) const = 0;
      //    Looks for the first object with the given key.
      //    Returns a pointer to this object if found and sets the 
      //    iterator to the position of that object. Repeatedly calling 
      //    this iterator findNext(key) will allow you to enumerate 
      //    all this container's objects that have the same key.
      //    Returns NULL and leaves the iterator unchanged if no object
      //    with this key was found.

      virtual MIRContainerItem* remove(Iterator* iterator) = 0;
      //    Removes the element designated by iterator.
      //    Returns a pointer to that object if successful. The iterator 
      //    then points to the element replacing the removed element or 
      //    to the last element in the container if the removed element 
      //    was the last one.
      //    If the operation fails it returns NULL and the iterator is 
      //    left unchanged.


    /*
     * Defines an abstract class for iterators on some kind of container.
     */
    class MIR_STD_API Iterator {

      public:

        //## Destruction

          virtual ~Iterator();
          //    Releases the resources used by this iterator.


        //## Query methods

          //virtual int operator==(const Iterator& right) const;
          //virtual int operator!=(const Iterator& right) const;
          //    Tests for (in)equality between two iterators.


        //## General functions

          //virtual operator=(const Iterator& right);
          //    Copies the iterator given as a parameter

          virtual MIRContainer* getContainer() const;
          //    Returns the container this iterator is operating on.

          virtual MIRContainerItem* operator*() const = 0;
          virtual MIRContainerItem* get() const = 0;
          //    Returns a pointer to the object for the current position
          //    of the iterator or NULL if the iterator is invalid.

        //## Methods for moving iterators

          virtual MIRContainerItem* moveToFirst() = 0;
          //    Moves the iterator to the first element of the container. 
          //    Returns a pointer to the first object if successful
          //    and NULL otherwise

          virtual MIRContainerItem* moveToLast() = 0;
          //    Moves the iterator to the last element of the container. 
          //    Returns a pointer to the last object if successful
          //    and NULL otherwise

          virtual MIRContainerItem* next(int nth=1) = 0;
          //    Moves the iterator to the nth element to the "right", starting 
          //    from the current position, and returns the corresponding object.
          //    If this object does not exist, because we reached the last
          //    element, the iterator points to the last element of the 
          //    container and NULL is returned.
          //
          //    Note:
          //     - we must have nth >= 0

          virtual MIRContainerItem* previous(int nth=1) = 0;
          //    Moves the iterator to the nth element to the "left", starting
          //    from the current position, and returns the corresponding object.
          //    If this object does not exist, because we reached the first
          //    element, the iterator points to the first element of the 
          //    container and NULL is returned.
          //
          //    Note:
          //     - we must have nth >= 0

          virtual MIRContainerItem* findFirst(const MIRContainerKey* key) = 0;
          //    Returns the first object with the given key in the container 
          //    and moves the iterator to the position of that object.
          //    If no such element can be found, NULL is returned and the 
          //    iterator is left unchanged.

          virtual MIRContainerItem* findLast(const MIRContainerKey* key) = 0;
          //    Returns the last object with the given key in the container
          //    and moves the iterator to the position of that object.
          //    If no such element can be found, NULL is returned and the 
          //    iterator is left unchanged.

          virtual MIRContainerItem* findNext(const MIRContainerKey* key) = 0;
          //    This method should only be called if you successfully called 
          //    one of the "find" operations on this iterator before.
          //
          //    Returns the next element with the given key starting from the 
          //    current position of the iterator and moves the iterator to the 
          //    position of that object.
          //    If no such element can be found, NULL is returned and the 
          //    iterator is left in a container dependent position.

          virtual MIRContainerItem* findPrevious(const MIRContainerKey* key) = 0;
          //    This method should only be called if you successfully called 
          //    one of the "find" operations on this iterator before.
          //
          //    Returns the previous element with the given key starting from the 
          //    current position of the iterator and moves the iterator to the 
          //    position of that object.
          //    If no such element can be found, NULL is returned and the 
          //    iterator is left in a container dependent position.

        protected:

          const MIRContainer* container;
    };

  protected:

    //## Ordering functions
      // Objects put in containers are supposed to have a key. 
      // The following compare functions define an order on these keys.
      // This allows searches to be performed and the implementation of 
      // containers that require that their elements be sorted.

      virtual int compare2Object(const MIRContainerItem* ptr1, const MIRContainerItem* ptr2) const;
      //    Compares the keys of two objects. This operation is used 
      //    when inserting/removing an object and in the contains function.
      //    Returns <0 if key of ptr1 < key of ptr2
      //             0 if key of ptr1 == key of ptr2
      //            >0 if key of ptr1 > key of ptr2
      //    By default this function compares the pointer values of 
      //    ptr1 and ptr2.

      virtual int compare2Key(const MIRContainerItem* ptr, const MIRContainerKey* key) const;
      //    Compares the key of ptr with the given key. This operation is used
      //    when in the functions that look for objects corresponding to a
      //    specific key. It is not used when inserting/removing an object or 
      //    in the contains function.
      //    Returns <0 if key of ptr1 < key
      //             0 if key of ptr1 == key
      //            >0 if key of ptr1 > key
      //    By default this function compares the pointer value of ptr
      //    with that of key.
};

inline int _MIRContainer::getSize() const
{
    return size;
}

typedef MIRContainer::Iterator MIRIterator;

inline _MIRContainer::_MIRContainer():
    size(0)
{
}

inline _MIRContainer::~_MIRContainer()
{
}

inline MIRContainer::MIRContainer():
    _MIRContainer()
{
}

inline MIRContainer::~MIRContainer()
{
}

inline int MIRContainer::compare2Object(const MIRContainerItem* ptr1, const MIRContainerItem* ptr2) const
{
    return (char*)ptr1-(char*)ptr2;
}

inline int MIRContainer::compare2Key(const MIRContainerItem* ptr, const MIRContainerKey* key) const
{
    return (char*)ptr-(char*)key;
}

inline MIRContainer* MIRContainer::Iterator::getContainer() const
{
    return (MIRContainer*)container;
}

inline MIRContainer::Iterator::~Iterator()
{
}

#endif /* end of file */

