//####################################################################
//##     Copyright  Meta Integration Technology, Inc. 1997-2002    ##
//##                      All Rights Reserved.                      ##
//####################################################################
//## Meta Integration(tm) Model Bridge (MIMB), Works (MIW),         ##
//## Repository (MIR), Repository SDK (MIRSDK), and Database SDK    ##
//## (MIDBSDK) are trademarks of Meta Integration Technology, Inc.  ##
//## Other product and company names (or logos) mentioned herein    ##
//## may be the trademarks of their respective owners.              ##
//####################################################################
//##                 http://www.metaintegration.com                 ##
//####################################################################


#ifndef MIRHashtable_h
#define MIRHashtable_h 1

/*
 * Implements the MIRContainer abstract class using a hashtable.
 *
 * Notes:
 * - This implementation does not preserve the order in which the 
 *   elements have been put in the container.
 * - The hashtable size is always rounded to the next power of 2.
 */

#include "MIRString.h"
#include "MIRContainer.h"

#define DEFAULT_SIZE 1024

class MIR_STD_API MIRHashtableElement;

class MIR_STD_API MIRHashtable: public MIRContainer
{
    
  public:

      class Iterator;
      friend class Iterator;

    //## Construction

      MIRHashtable(int _size=DEFAULT_SIZE);
      //    Creates an empty hashtable. Size is the size of the array to 
      //    allocate when the first hashtable element is added.

      MIRHashtable(const MIRHashtable& src, int _size=-1);
      //    Creates a new hashtable initialising it with the content of 
      //    the given source hashtable. Note that only the object 
      //    pointers are copied, the objects themselves are not duplicated.
      //    If size is -1 then the new hashtable has the same size as src
      //    otherwise size specifies the size of the new hashtable.

      static void* operator new(size_t size);
      static void operator delete(void* ptr);
#ifdef WIN32
      static void* operator new(size_t size, const char* file, int line);
      static void* operator new(size_t size, int block_type, const char* file, int line);
#if _MSC_VER >= 1200
      static void operator delete(void* ptr, const char* file, int line);
      static void operator delete(void* ptr, int block_type, const char* file, int line);
#endif
#endif
      //    Used to allocate/deallocate hashtables (and subclasses). 

      virtual MIRHashtable& operator=(const MIRHashtable& right);
      //    Replaces the content of the current hashtable with that of the
      //    given hashtable. Note that only the object pointers are copied,
      //    the objects themselves are not duplicated.


    //## Destruction

      virtual ~MIRHashtable();
      //    Releases the resources used by this hashtable. Note that this 
      //    operation does not release the memory used by the objects 
      //    that have been put in the hashtable.


    //## Query methods

      virtual int operator==(const MIRHashtable& right) const;
      virtual int operator!=(const MIRHashtable& right) const;
      //    Tests for (in)equality of the contents of two hashtables
      //    by comparing the values of the pointers they contain.


    //## General functions

      virtual MIRBoolean empty();
      //    Removes all the elements that were in the hashtable.
      //    Returns true if successful and false otherwise.

      virtual MIRBoolean resize(int _size);
      //    Resizes the hash table. The hashtable size is rounded up
      //    to the next power of 2. Note that this is a long operation
      //    and should not be done too often.


    //## Creating and deleting an iterator for this container

      virtual MIRIterator* newIterator(MIRBoolean first=true) const;
      //    Allocates an iterator operating on this container. Note 
      //    that an iterator is specific to the container type that 
      //    allocated it.
      //    If first is true then the iterator is initialised so that it 
      //    points to the first element in the container. Otherwise it 
      //    designates the last element of the container.
      //    Returns a pointer to the new iterator if successful and
      //    NULL otherwise.

      virtual void deleteIterator(MIRIterator* iterator) const;
      //    Releases an iterator.


    //## Iterator based methods for adding and removing elements
    //## (Container independant)

      virtual MIRBoolean add(const MIRContainerItem* ptr, MIRBoolean unique=true, MIRIterator* iterator=NULL);
      //    Adds a new object to the container. If unique is true then
      //    the insertion will fail is an object with the same key
      //    is already in the container.
      //    If the insertion succeeds and iterator is not NULL then it is
      //    set to the position where the object has been inserted and true
      //    is returned. Otherwise, false is returned and iterator is left
      //    unchanged.

      virtual MIRBoolean contains(const MIRContainerItem* ptr, MIRIterator* iterator=NULL) const;
      //    Returns true if the container contains the object
      //    pointed to by "ptr" and sets the iterator to the position
      //    of this object. Returns false and leaves the iterator 
      //    unchanged otherwise.

      virtual MIRContainerItem* findFirst(const MIRContainerKey* key, MIRIterator* iterator=NULL) const;
      //    Looks for the first object with the given key.
      //    Returns a pointer to this object if found and sets the 
      //    iterator to the position of that object. Repeatedly calling 
      //    this iterator nextElement(key) will allow you to enumerate 
      //    all this container's objects that have the same key.
      //    Returns NULL and leaves the iterator unchanged if no object
      //    with this key was found.

      virtual MIRContainerItem* findLast(const MIRContainerKey* key, MIRIterator* iterator=NULL) const;
      //    Looks for the last object with the given key.
      //    Returns a pointer to this object if found and sets the 
      //    iterator to the position of that object. Repeatedly calling 
      //    this iterator previousElement(key) will allow you to enumerate 
      //    all this container's objects that have the same key.
      //    Returns NULL and leaves the iterator unchanged if no object
      //    with this key was found.

      virtual MIRContainerItem* remove(MIRIterator* iterator);
      //    Removes the element designated by iterator.
      //    Returns a pointer to that object if successful. The iterator 
      //    then points to the element replacing the removed element or 
      //    to the last element in the container if the removed element 
      //    was the last one.
      //    If the operation fails it returns NULL and the iterator is 
      //    left unchanged.


      class MIR_STD_API Iterator: public MIRContainer::Iterator {

      public:

        friend class MIRHashtable;

        //## Construction

          Iterator();
          //    Creates a new iterator not associated to a container.

          Iterator(const MIRHashtable& htable, MIRBoolean first=true);
          //    Creates a new iterator pointing to the first element 
          //    in the hash table if first is true or the last otherwise.

          static void* operator new(size_t size);
          static void operator delete(void* ptr);
#ifdef WIN32
          static void* operator new(size_t size, const char* file, int line);
          static void* operator new(size_t size, int block_type, const char* file, int line);
#if _MSC_VER >= 1200
          static void operator delete(void* ptr, const char* file, int line);
          static void operator delete(void* ptr, int block_type, const char* file, int line);
#endif
#endif
          //    Used to allocate/deallocate iterators (and subclasses). 


        //## Destruction

          virtual ~Iterator();
          //    Releases the resources used by this iterator.


        //## Query methods

          virtual int operator==(const Iterator& right) const;
          virtual int operator!=(const Iterator& right) const;
          //    Tests for (in)equality between two iterators.


        //## General functions

          virtual Iterator& operator=(const Iterator& right);
          //    Copies the iterator given as a parameter

          virtual MIRContainerItem* operator*() const;
          virtual MIRContainerItem* get() const;
          //    Returns a pointer to the object for the current position
          //    of the iterator or NULL if the iterator is invalid.


        //## Methods for moving iterators

          virtual MIRContainerItem* moveTo(MIRHashtable& htable, MIRBoolean first=true);
          //    Moves the iterator to the specified container.
          //    If first is true then the iterator is initialised so that it 
          //    points to the first element in the container. Otherwise it 
          //    designates the last element of the container.
          //    Returns a pointer to the new iterator if successful and
          //    NULL otherwise.

          virtual MIRContainerItem* moveToFirst();
          //    Moves the iterator to the first element of the container. 
          //    Returns a pointer to the first object if successful
          //    and NULL otherwise

          virtual MIRContainerItem* moveToLast();
          //    Moves the iterator to the last element of the container. 
          //    Returns a pointer to the last object if successful
          //    and NULL otherwise

          virtual MIRContainerItem* next(int nth=1);
          //    Moves the iterator to the nth element to the "right", starting 
          //    from the current position, and returns the corresponding object.
          //    If this object does not exist, because we reached the last
          //    element, the iterator points to the last element of the 
          //    container and NULL is returned.

          virtual MIRContainerItem* previous(int nth=1);
          //    Moves the iterator to the nth element to the "left", starting
          //    from the current position, and returns the corresponding object.
          //    If this object does not exist, because we reached the first
          //    element, the iterator points to the first element of the 
          //    container and NULL is returned.

          virtual MIRContainerItem* findFirst(const MIRContainerKey* key);
          //    Returns the first object with the given key in the container 
          //    and moves the iterator to the position of that object.
          //    If no such element can be found, NULL is returned and the 
          //    iterator is left unchanged.

          virtual MIRContainerItem* findLast(const MIRContainerKey* key);
          //    Returns the last object with the given key in the container
          //    and moves the iterator to the position of that object.
          //    If no such element can be found, NULL is returned and the 
          //    iterator is left unchanged.

          virtual MIRContainerItem* findNext(const MIRContainerKey* key);
          //    This method should only be called if you successfully called 
          //    one of the "find" operations on this iterator before.
          //
          //    Returns the next element with the given key starting from the 
          //    current position of the iterator and moves the iterator to the 
          //    position of that object.
          //    If no such element can be found, NULL is returned and the 
          //    iterator points to the last element with the given key.

          virtual MIRContainerItem* findPrevious(const MIRContainerKey* key);
          //    This method should only be called if you successfully called 
          //    one of the "find" operations on this iterator before.
          //
          //    Returns the previous element with the given key starting from the 
          //    current position of the iterator and moves the iterator to the 
          //    position of that object.
          //    If no such element can be found, NULL is returned and the 
          //    iterator points to the first element with the given key.


      protected:

          MIRBoolean _moveTo(int vpos);

          const MIRHashtableElement* element;
          int index;
    };


  protected:

    //## Hashing functions
      // To put an object in the hashtable you must be able to compute a hash value
      // based on the key of that object.

      virtual unsigned int hashObject(const MIRContainerItem* ptr) const;
      //    Returns a hash value based on the object pointed to by ptr. 
      //    This value is a 32 bit unsigned integer which is later warped 
      //    to the size of the hashtable.
      //    The default implementation just returns the value of ptr.

      virtual unsigned int hashKey(const MIRContainerKey* key) const;
      //    Returns a hash value based on the key pointed to by ptr. This 
      //    value is a 32 bit unsigned integer which is later warped to the 
      //    size of the hashtable.
      //    The default implementation just returns the value of the key 
      //    pointer.


      MIRHashtableElement* *table;
      int table_mask;
};

typedef MIRHashtable::Iterator MIRHashtableIterator;

class MIR_STD_API MIRHashtableElement
{

      friend class MIRHashtable;
      friend class MIRHashtable::Iterator;

  private:
      MIRHashtableElement();
      MIRHashtableElement(MIRHashtableElement* _previous, MIRHashtableElement* _next, MIRContainerItem* _data);
      static void* operator new(size_t size);
      static void operator delete(void* ptr);
#ifdef WIN32
      static void* operator new(size_t size, const char* file, int line);
      static void* operator new(size_t size, int block_type, const char* file, int line);
#if _MSC_VER >= 1200
      static void operator delete(void* ptr, const char* file, int line);
      static void operator delete(void* ptr, int block_type, const char* file, int line);
#endif
#endif

      MIRHashtableElement* next;
      MIRHashtableElement* previous;
      MIRContainerItem* data;
};



/*
 * Inline methods
 */

#ifdef WIN32
inline void* MIRHashtable::operator new(size_t size, const char* file, int line)
{
    return operator new(size,_NORMAL_BLOCK,file,line);
}

#if _MSC_VER >= 1200
inline void MIRHashtable::operator delete(void* ptr, int block_type, const char* file, int line)
{
    operator delete(ptr);
}

inline void MIRHashtable::operator delete(void* ptr, const char* file, int line)
{
    operator delete(ptr);
}
#endif
#endif

inline MIRIterator* MIRHashtable::newIterator(MIRBoolean first) const
{
    return _new_dbg MIRHashtable::Iterator(*this,first);
}

inline void MIRHashtable::deleteIterator(MIRIterator* iterator) const
{
    delete iterator;
}

inline unsigned int MIRHashtable::hashObject(const MIRContainerItem* ptr) const
{
    return (unsigned int)ptr;
}

inline unsigned int MIRHashtable::hashKey(const MIRContainerKey* key) const
{
    return (unsigned int)key;
}

#ifdef WIN32
inline void* MIRHashtable::Iterator::operator new(size_t size, const char* file, int line)
{
    return operator new(size,_NORMAL_BLOCK,file,line);
}

#if _MSC_VER >= 1200
inline void MIRHashtable::Iterator::operator delete(void* ptr, int block_type, const char* file, int line)
{
    operator delete(ptr);
}

inline void MIRHashtable::Iterator::operator delete(void* ptr, const char* file, int line)
{
    operator delete(ptr);
}
#endif
#endif

inline MIRHashtable::Iterator::~Iterator()
{
}

inline MIRHashtableElement::MIRHashtableElement()
{
}

inline MIRHashtableElement::MIRHashtableElement(MIRHashtableElement* _previous, MIRHashtableElement* _next, MIRContainerItem* _data):
    next(_next),previous(_previous),data(_data)
{
}

#ifdef WIN32
inline void* MIRHashtableElement::operator new(size_t size, const char* file, int line)
{
    return operator new(size,_NORMAL_BLOCK,file,line);
}

#if _MSC_VER >= 1200
inline void MIRHashtableElement::operator delete(void* ptr, int block_type, const char* file, int line)
{
    operator delete(ptr);
}

inline void MIRHashtableElement::operator delete(void* ptr, const char* file, int line)
{
    operator delete(ptr);
}
#endif
#endif


#endif /* end of file */
