//####################################################################
//##     Copyright  Meta Integration Technology, Inc. 1997-2002    ##
//##                      All Rights Reserved.                      ##
//####################################################################
//## Meta Integration(tm) Model Bridge (MIMB), Works (MIW),         ##
//## Repository (MIR), Repository SDK (MIRSDK), and Database SDK    ##
//## (MIDBSDK) are trademarks of Meta Integration Technology, Inc.  ##
//## Other product and company names (or logos) mentioned herein    ##
//## may be the trademarks of their respective owners.              ##
//####################################################################
//##                 http://www.metaintegration.com                 ##
//####################################################################


#ifndef MIRMessagesObject_h
#define MIRMessagesObject_h 1

#include "MIRString.h"
#include "MIRMessage.h"
#include "MIRVector.h"
#include "MIRStdDefs.h"

#ifndef NO_IOSTREAM
#include "MIRIO.h"
#define MIR_OSTREAM ostream
#else
class mir_ostream;
#define MIR_OSTREAM mir_ostream
#endif

#define MIR_STD_OUT (MIR_OSTREAM*) -1
#define MIR_STD_ERR (MIR_OSTREAM*) -2
/*
 * Defines message types.
 */
enum MIRMessageType
{
    MIR_INFO = 0,       // e.g. copyright, version, ....
    MIR_STATUS,         // progression messages
    MIR_WARNING,        // e.g. not yet implemented features
    MIR_ERROR,          // error but execution continues
    MIR_FATAL,          // unrecoverable error
    MIR_DEBUG,
	MIR_NONE            // empty message type
};

class MIR_STD_API MIRMessagesFormat
{
public:
    MIRMessagesFormat(const char* name, MIRMessage* messages, int offset, int count);
    virtual ~MIRMessagesFormat();
    MIRMessage* getMessage(unsigned int id);
    const char* getMessageString(unsigned int id);
private:
    const char* name;
    MIRMessage* messages;
    int offset,count;
};

class MIR_STD_API MIRI18nDynamicMessagesFormat: public MIRMessagesFormat
{
public:
    MIRI18nDynamicMessagesFormat(int& status, const char* name, MIRMessage* messages, int offset, int count);
};

/*
 * MIRLog implements the MIR logging mechanism used 
 * to log bridge informations and errors. It thus provides 
 * methods to send the log to a stream, filter the messages and
 * send messages to the log.
 */
class MIR_STD_API MIRLog
{
public:

    //## Construction

    MIRLog();
    //    Initialises a MessageObject object with no output stream

    MIRLog(MIR_OSTREAM* os);
    //    Initialises a MessageObject object with the specified output stream

    static void* operator new(size_t size);
    static void operator delete(void* ptr);
#ifdef WIN32
    static void* operator new(size_t size, const char* file, int line);
    static void* operator new(size_t size, int block_type, const char* file, int line);
#if _MSC_VER >= 1200
    static void operator delete(void* ptr, const char* file, int line);
    static void operator delete(void* ptr, int block_type, const char* file, int line);
#endif
#endif
    //    Used to allocate/deallocate MIRObject objects (and subclasses). 


    //## Destruction

    virtual ~MIRLog();
    //    Releases the resources used by this object. Note that this 
    //    does not close the output stream.


    //## General methods
    void setEncoding(const MIRTString& encoding);
    void setStream(MIR_OSTREAM* os);
    void setTimestamp(MIRBoolean timestamp);
    void setAppendEOL(MIRBoolean append);

    const MIRString& getLog() const;
    void clearLog();

    void printMessage(const MIRMessageType sType, const int nNumber,
      const char* arg1=NULL,  const char* arg2=NULL,  const char* arg3=NULL,
      const char* arg4=NULL,  const char* arg5=NULL,  const char* arg6=NULL,
      const char* arg7=NULL,  const char* arg8=NULL,  const char* arg9=NULL,
      const char* arg10=NULL, const char* arg11=NULL, const char* arg12=NULL);
    void printMessage(const MIRMessageType sType, const MIRString& sMessage);
    //    Print the specified message to the output stream. If the 
    //    stream is NULL do nothing.

    //## Methods for throwing exceptions
    void ThrowException(const int nNumber,
      const char* arg1=NULL,  const char* arg2=NULL,  const char* arg3=NULL,
      const char* arg4=NULL,  const char* arg5=NULL,  const char* arg6=NULL,
      const char* arg7=NULL,  const char* arg8=NULL,  const char* arg9=NULL,
      const char* arg10=NULL, const char* arg11=NULL, const char* arg12=NULL);
    void ThrowException(const MIRString& sMessage);
    //    these functions do the following steps:
    //    create an exception object and set its error number
    //    build the message string from the parameters and set the exception object with it
    //    throw the exception.


protected:
    // Different encoders needed:
    //   data for the data passed as arguments of printMessage and ThrowException
    //   utf8 for the message itself
    //   out  for printing the message into the stream (NOT IMPLEMENTED YET)
    MIREncodingConverter* data;
    MIREncodingConverter* utf8;
    MIREncodingConverter* out;
    int err,cnt;

    // The messages holder
    MIRString sMessages;
    MIRBoolean appendEOL;
    MIR_OSTREAM* pLogStream;

    MIRBoolean printTimestamp;

    void processArguments(MIRTString tmp[12], int& argCount, int& argLength, const char* arg1, ...);
    MIRTString buildMessage(const MIRMessageType sType, const char* sMessage, int messageID,...);
    //    Function used to build the message.

    virtual void log(const MIRTString& message);
};

extern MIR_STD_API MIRLog MirStdOut;
extern MIR_STD_API MIRLog MirStdErr;

/*
 * Inline methods
 */

#ifdef WIN32
inline void* MIRLog::operator new(size_t size, const char* file, int line)
{
    return operator new(size,_NORMAL_BLOCK,file,line);
}

#if _MSC_VER >= 1200
inline void MIRLog::operator delete(void* ptr, int block_type, const char* file, int line)
{
    operator delete(ptr);
}

inline void MIRLog::operator delete(void* ptr, const char* file, int line)
{
    operator delete(ptr);
}
#endif
#endif

inline const MIRString& MIRLog::getLog() const
{
    return sMessages;
}

inline void MIRLog::clearLog()
{
    sMessages.empty();
}

inline void MIRLog::setStream(MIR_OSTREAM* pStream)
{
    pLogStream = pStream;
}

inline void MIRLog::setTimestamp(MIRBoolean timestamp)
{
    printTimestamp = timestamp;
}

inline void MIRLog::setAppendEOL(MIRBoolean append)
{
    appendEOL = append;
}

#endif /* end of file */
