//####################################################################
//##     Copyright  Meta Integration Technology, Inc. 1997-2003    ##
//##                      All Rights Reserved.                      ##
//####################################################################
//## Meta Integration(tm) Model Bridge (MIMB), Works (MIW),         ##
//## Repository (MIR), Repository SDK (MIRSDK), and Database SDK    ##
//## (MIDBSDK) are trademarks of Meta Integration Technology, Inc.  ##
//## Other product and company names (or logos) mentioned herein    ##
//## may be the trademarks of their respective owners.              ##
//####################################################################
//##                 http://www.metaintegration.com                 ##
//####################################################################



#ifndef    MIROption_h
#define    MIROption_h 1


// ## Define 

#include "MIRStdDefs.h"
#include "MIRString.h"
#include "MIRStringList.h"

/*
 * Predefined values or structures that may be used by the bridges 
 */


// List of possible value for a bridge option type

#define OPTION_TYPE_STRING              "String"
#define OPTION_TYPE_ENUMERATED          "Enumerated"
#define OPTION_TYPE_NUMERIC             "Numeric"
#define OPTION_TYPE_BOOLEAN             "Boolean"
#define OPTION_TYPE_PASSWORD            "Password"
#define OPTION_TYPE_FILE                "File"
#define OPTION_TYPE_DIRECTORY           "Directory"
#define OPTION_TYPE_REPOSITORY_MODEL    "Repository_Model"
#define OPTION_TYPE_REPOSITORY_SUBSET   "Repository_Subset"

// This private class is used in the implementation of the option
// classes and thus must be declared here but it is not part of the 
// official API.
class MIROptionInfo;

/*
 * Class that implements the Option
 */

class MIR_STD_API MIROption
{
public:

    //## Construction

      MIROption() {};
      MIROption(const MIRTString& name);

      MIROption(MIROptionInfo *info);
      //    This constructor is defined for internal use only.
      //    Although it is public (for implementation reasons),
      //    it is not part of the official API.


    //## Destruction

      virtual ~MIROption();


   //## Name Attribute Operations

      virtual void setName(const MIRTString& value);
      virtual const MIRString& getName() const;
      //    Defines the name attribute of the option.
      //    An option is identified by its name.

   //## Display Name Attribute Operations

      virtual void setDisplayName(const MIRTString& value);
      virtual const MIRString& getDisplayName() const;
      //    Defines the display name attribute of the option.
      //    This is a localized version of the name of the option.


   //## Type Attribute Operations

      virtual void setType(const MIRTString& value);
      virtual const MIRString& getType() const;
      //    Defines the type attribute of the option.


   //## DefaultValue Attribute Operations

      virtual void setDefaultValue(const MIRTString& value);
      virtual const MIRString& getDefaultValue() const;
      virtual const MIRString& getDisplayDefaultValue() const;
      //    Defines the defaultValue attribute of the option.
      //    You can retrieve either the default value itself or 
      //    its display value (a loacalized version of the value)
      //    if the type is Enumerated or Boolean.


   //## Description Attribute Operations

      virtual void setDescription(const MIRTString& value);
      virtual const MIRString& getDescription() const;
      //    Defines the description attribute of the option.

	  
   //## Value Attribute Operations

      virtual void setValue(const MIRTString& value);
      virtual const MIRString& getSpecifiedValue() const;
      //    Defines the Value attribute of the option.
      virtual const MIRString& getValue() const;
      //    Return Value if it exists otherwise return the DefaultValue
      virtual const MIRString& getDisplayValue() const;
      //    Return Value if it exists otherwise return the DefaultValue
      //    If the type is Enumerated or Boolean, it returns the localized
      //    version of the value instead.


   //## PossibleValues Attribute Operations

      MIRBoolean addPossibleValue(const MIRString& value);
      MIRBoolean addPossibleValue(const MIRString& value, const MIRString& displayValue);
      //    Associates the PossibleValue designated by the given value to this Option.
      //    Returns TRUE if successful, or FALSE otherwise (if it exists).

      int getPossibleValueCount() const;
      //    Returns the number of PossibleValue(s) associated to this Option.

      const MIRString& getPossibleValue(int position) const;
      const MIRString& getDisplayPossibleValue(int position) const;
      //    Returns the PossibleValue associated to this Option
      //    located at the given position,
      //    Returns empty if not successful.

      const MIRString& getDisplayPossibleValue(const MIRString& value) const;
      //    Returns the PossibleValue display value associated to this Option
      //    given its actual value,
      //    Returns empty if not successful.

      MIRBoolean removePossibleValue(int position);
      //    Removes the PossibleValue located at the given position.
      //    Returns TRUE if successful, or FALSE otherwise.

      void removePossibleValues();
      //    Removes all PossibleValues

	  MIRBoolean valueIsPossibleValue (const MIRString& value) const;
      //    Check if value is a PossibleValue.
      //    Returns TRUE if successful, or FALSE otherwise (if it exists).

   //## Mandatory Attribute Operations

      virtual void setMandatory(MIRBoolean value);
      virtual MIRBoolean isMandatory() const;
      //    If the option is mandatory set to TRUE, otherwise to FALSE
      //    It is FALSE by default.

protected:

    //## Attributes

       MIRString aName;
       MIRString aDisplayName;
       MIRString aType;
       MIRString aDefaultValue;
       MIRString aDescription;
       MIRString aValue;
       MIRBoolean aMandatory;

       MIRStringList possibleValues;
       MIRStringList displayPossibleValues;
};


/*
 * MIROption Inline methods
 */

inline void MIROption::setName(const MIRTString& value)
{
    aName=value;
}

inline const MIRString& MIROption::getName() const
{
    return aName;
}

inline void MIROption::setDisplayName(const MIRTString& value)
{
    aDisplayName=value;
}

inline const MIRString& MIROption::getDisplayName() const
{
    return aDisplayName.isEmpty()?aName:aDisplayName;
}

inline void MIROption::setType(const MIRTString& value)
{
    aType=value;
}

inline const MIRString& MIROption::getType() const
{
    return aType;
}

inline void MIROption::setDescription(const MIRTString& value)
{
    aDescription=value;
}

inline const MIRString& MIROption::getDescription() const
{
    return aDescription;
}

inline void MIROption::setDefaultValue(const MIRTString& value)
{
    aDefaultValue=value;
}

inline const MIRString& MIROption::getDefaultValue() const
{
    return aDefaultValue;
}

inline void MIROption::setValue(const MIRTString& value)
{
    aValue=value;
}

inline const MIRString& MIROption::getSpecifiedValue() const
{
    return aValue;
}

inline int MIROption::getPossibleValueCount() const
{
    return possibleValues.getSize();
}

inline void MIROption::setMandatory(MIRBoolean value)
{
    aMandatory = value;
}

inline MIRBoolean MIROption::isMandatory() const
{
    return aMandatory;
}


/*
 * MIROptionVector
 *
 * Notes:
 *  - MIROptionVector enforces name uniqueness.
 */
class MIR_STD_API MIROptionVector: public MIROrderedVector
{
  public:

    //## Construction

      MIROptionVector();
      //    Creates an empty nary link.


    //## Ordering functions -> based on names

      virtual int compare2Object(const MIRContainerItem* ptr1, const MIRContainerItem* ptr2) const;
      //    Compares the keys of two objects.
      //    Returns <0 if ptr1.key < ptr2.key
      //             0 if ptr1.key == ptr2.key
      //            >0 if ptr1.key > ptr2.key

      virtual int compare2Key(const MIRContainerItem* ptr, const MIRContainerKey* key) const;
      //    Compares the key of ptr with key.
      //    Returns <0 if ptr1.key < key
      //             0 if ptr1.key == key
      //            >0 if ptr1.key > key
};

typedef MIRVector::Iterator MIROptionVectorIterator;


/*
 * Class that implements the  Option List
 */
class MIR_STD_API MIROptionList
{
  public:

    //## Construction

      MIROptionList() {};

      void initialize(MIROptionInfo *options);
      //    This constructor is defined for internal implementation
      //    only. Although it is public (for implementation reasons),
      //    it is not part of the official API.

 

    //## Destruction

      virtual ~MIROptionList();


    //## Option  Operations

      MIRBoolean add(MIROption* option);
      //    Associates the Option designated by the given value to this list.
      //    Returns TRUE if successful, or FALSE otherwise (if it exists).

      int getCount() const;
      //    Returns the number of Option(s) associated to this list.

      MIRBoolean contains(const MIROption* option) const;
      //    Returns TRUE if option is already in the list.
      //    Returns FALSE otherwise.

      MIROption* get(int position) const;
      //    Returns the Option associated to this list
      //    located at the given position,
      //    Returns NULL if not successful.

      MIROption* get(const MIRString& name) const;
      //    Returns the first Option associated to this list
      //    designated by the given name. 
      //    Returns NULL if not successful.

      MIROption* remove(int position);
      //    Removes the Option located at the given position.
      //    Returns NULL if not successful.
 
      MIROption* remove(const MIRString& name);
      //    Removes the Option designated by the given name.
      //    Returns NULL if not successful.

      MIROption* remove(MIROption* option);
      //    Removes the Option designated by the given value.
      //    Returns NULL if not successful.

      void removeAll();
      //    Removes and deletes all the Options.

      const MIRString& getValue(const MIRString& name) const;
      //    Returns the value corresponding to the Option designated
      //    by the given name
      //    returns emptyString if the option doesn't exist


protected:
    //## Attributes

      MIROptionVector options;

};

//## Simple nary links operations (inline)

inline int MIROptionList::getCount() const
{
    return options.getSize();
}

inline MIRBoolean MIROptionList::contains(const MIROption* option) const
{
    return options.contains(option);
}

inline MIROption* MIROptionList::get(int position) const
{
    return (MIROption*)(options)[position];
}

inline MIROption* MIROptionList::get(const MIRString &name) const
{
    return (MIROption*)options.findFirst(&name);
}

inline MIRBoolean MIROptionList::add(MIROption* option)
{
    return options.add(option, true);
}

inline MIROption *MIROptionList::remove(int position)
{
    return (MIROption *)options.removeElement(position);
}




#endif /* end of file */
