//####################################################################
//##     Copyright  Meta Integration Technology, Inc. 1997-2002    ##
//##                      All Rights Reserved.                      ##
//####################################################################
//## Meta Integration(tm) Model Bridge (MIMB), Works (MIW),         ##
//## Repository (MIR), Repository SDK (MIRSDK), and Database SDK    ##
//## (MIDBSDK) are trademarks of Meta Integration Technology, Inc.  ##
//## Other product and company names (or logos) mentioned herein    ##
//## may be the trademarks of their respective owners.              ##
//####################################################################
//##                 http://www.metaintegration.com                 ##
//####################################################################


#ifndef MIRRBTree_h
#define MIRRBTree_h 1

/*
 * Implements the MIRContainer abstract class using a red-black tree.
 *
 * Notes:
 * - This implementation does not preserve the order in which the 
 *   elements have been put in the container.
 */

#include "MIRString.h"
#include "MIRContainer.h"

class MIR_STD_API MIRRBTreeNode;

class MIR_STD_API MIRRBTree: public MIRContainer
{
    
  public:

      class Iterator;
      friend class Iterator;

    //## Construction

      MIRRBTree();
      //    Creates an empty binary tree.

      MIRRBTree(const MIRRBTree& src);
      //    Creates a new binary tree initialising it with the content of 
      //    the given source binary tree. Note that only the object 
      //    pointers are copied, the objects themselves are not duplicated.

      static void* operator new(size_t size);
      static void operator delete(void* ptr);
#ifdef WIN32
      static void* operator new(size_t size, const char* file, int line);
      static void* operator new(size_t size, int block_type, const char* file, int line);
#if _MSC_VER >= 1200
      static void operator delete(void* ptr, const char* file, int line);
      static void operator delete(void* ptr, int block_type, const char* file, int line);
#endif
#endif
      //    Used to allocate/deallocate trees (and subclasses). 

      virtual MIRRBTree& operator=(const MIRRBTree& right);
      //    Replaces the content of the current binary tree with that of the
      //    given binary tree. Note that only the object pointers are copied,
      //    the objects themselves are not duplicated.


    //## Destruction

      virtual ~MIRRBTree();
      //    Releases the resources used by this binary tree. Note that this 
      //    operation does not release the memory used by the objects 
      //    that have been put in the binary tree.


    //## Query methods

      virtual int operator==(const MIRRBTree& right) const;
      virtual int operator!=(const MIRRBTree& right) const;
      //    Tests for (in)equality of the contents of two binary trees
      //    by comparing the values of the pointers they contain.


    //## General functions

      virtual MIRBoolean empty();
      //    Removes all the elements that were in the binary tree.
      //    Returns true if successful and false otherwise.

      virtual void dump();
      //    Dumps the tree structure.


    //## Creating and deleting an iterator for this container

      virtual MIRIterator* newIterator(MIRBoolean first=true) const;
      //    Allocates an iterator operating on this container. Note 
      //    that an iterator is specific to the container type that 
      //    allocated it.
      //    If first is true then the iterator is initialised so that it 
      //    points to the first element in the container. Otherwise it 
      //    designates the last element of the container.
      //    Returns a pointer to the new iterator if successful and
      //    NULL otherwise.

      virtual void deleteIterator(MIRIterator* iterator) const;
      //    Releases an iterator.


    //## Iterator based methods for adding and removing elements
    //## (Container independant)

      virtual MIRBoolean add(const MIRContainerItem* ptr, MIRBoolean unique=true, MIRIterator* iterator=NULL);
      //    Adds a new object to the container. If unique is true then
      //    the insertion will fail is an object with the same key
      //    is already in the container.
      //    If the insertion succeeds and iterator is not NULL then it is
      //    set to the position where the object has been inserted and true
      //    is returned. Otherwise, false is returned and iterator is left
      //    unchanged.

      virtual MIRBoolean contains(const MIRContainerItem* ptr, MIRIterator* iterator=NULL) const;
      //    Returns true if the container contains the object
      //    pointed to by "ptr" and sets the iterator to the position
      //    of this object. Returns false and leaves the iterator 
      //    unchanged otherwise.

      virtual MIRContainerItem* findFirst(const MIRContainerKey* key, MIRIterator* iterator=NULL) const;
      //    Looks for the first object with the given key.
      //    Returns a pointer to this object if found and sets the 
      //    iterator to the position of that object. Repeatedly calling 
      //    this iterator nextElement(key) will allow you to enumerate 
      //    all this container's objects that have the same key.
      //    Returns NULL and leaves the iterator unchanged if no object
      //    with this key was found.

      virtual MIRContainerItem* remove(MIRIterator* iterator);
      //    Removes the element designated by iterator.
      //    Returns a pointer to that object if successful. The iterator 
      //    then points to the element replacing the removed element or 
      //    to the last element in the container if the removed element 
      //    was the last one.
      //    If the operation fails it returns NULL and the iterator is 
      //    left unchanged.


    //## Node based methods for adding and removing elements
    //## (Specific to red-black trees)

      virtual MIRRBTreeNode* addNode(const MIRContainerItem* ptr, MIRBoolean unique=true);
      //    Adds a new object to the container. If unique is true then
      //    the insertion will fail if an object with the same key
      //    is already in the container.
      //    Returns a pointer to the node if the insertion succeeds
      //    and NULL otherwise.

      virtual MIRRBTreeNode* getNode(const MIRContainerItem* ptr) const;
      //    Checks if the container contains the object pointed to by 
      //    ptr. If yes then returns a pointer to the node containing 
      //    that object. Returns NULL otherwise.

      virtual MIRRBTreeNode* firstNode() const;
      //    Returns a pointer to the node of the first object in the tree
      //    if successful. Returns NULL otherwise.

      virtual MIRRBTreeNode* lastNode() const;
      //    Returns a pointer to the node of the last object in the tree
      //    if successful. Returns NULL otherwise.

      virtual MIRRBTreeNode* nextNode(const MIRRBTreeNode* node) const;
      //    Returns a pointer to the node following the given node.
      //    Returns NULL if there is no such node.

      virtual MIRRBTreeNode* previousNode(const MIRRBTreeNode* node) const;
      //    Returns a pointer to the node preceding the given node.
      //    Returns NULL if there is no such node.

      virtual MIRRBTreeNode* findFirstNode(const MIRContainerKey* key) const;
      //    Retrieves the first object in the tree with the given key. 
      //    Returns a pointer to the node of this object if successful 
      //    and NULL otherwise. 

      virtual MIRRBTreeNode* findLastNode(const MIRContainerKey* key) const;
      //    Retrieves the last object in the tree with the given key. 
      //    Returns a pointer to the node of this object if successful 
      //    and NULL otherwise. 

      virtual MIRRBTreeNode* findNextNode(const MIRContainerKey* key, const MIRRBTreeNode* node) const;
      //    Looks for an object with the given key starting from
      //    the specified node. Returns a pointer to the node of 
      //    this object if successful and NULL otherwise.

      virtual MIRRBTreeNode* findPreviousNode(const MIRContainerKey* key, const MIRRBTreeNode* node) const;
      //    Looks for an object with the given key starting from
      //    the specified node. Returns a pointer to the node of 
      //    this object if successful and NULL otherwise.

      virtual MIRContainerItem* removeNode(MIRRBTreeNode* node);
      //    Removes the object contained in the specified node.
      //    Returns a pointer to that object if successful and NULL 
      //    otherwise.

      class MIR_STD_API Iterator: public MIRContainer::Iterator {

      public:

        friend class MIRRBTree;

        //## Construction

          Iterator();
          //    Creates a new iterator not associated to a container.

          Iterator(const MIRRBTree& tree, MIRBoolean first=true);
          //    Creates a new iterator pointing to the first element 
          //    in the tree if first is true or the last otherwise.

          Iterator(const MIRRBTree& tree, const MIRRBTreeNode& node);
          //    Creates an iterator corresponding to the specified
          //    tree node.

          static void* operator new(size_t size);
          static void operator delete(void* ptr);
#ifdef WIN32
          static void* operator new(size_t size, const char* file, int line);
          static void* operator new(size_t size, int block_type, const char* file, int line);
#if _MSC_VER >= 1200
          static void operator delete(void* ptr, const char* file, int line);
          static void operator delete(void* ptr, int block_type, const char* file, int line);
#endif
#endif
          //    Used to allocate/deallocate iterators (and subclasses). 


        //## Destruction

          virtual ~Iterator();
          //    Releases the resources used by this iterator.


        //## Query methods

          virtual int operator==(const Iterator& right) const;
          virtual int operator!=(const Iterator& right) const;
          //    Tests for (in)equality between two iterators.


        //## General functions

          virtual Iterator& operator=(const Iterator& right);
          //    Copies the iterator given as a parameter

          virtual MIRContainerItem* operator*() const;
          virtual MIRContainerItem* get() const;
          //    Returns a pointer to the object for the current position
          //    of the iterator or NULL if the iterator is invalid.

          const MIRRBTreeNode* getNode() const;
          //    Returns the node this iterator points to.


        //## Methods for moving iterators

          virtual MIRContainerItem* moveTo(MIRRBTree& tree, MIRBoolean first=true);
          //    Moves the iterator to the specified container.
          //    If first is true then the iterator is initialised so that it 
          //    points to the first element in the container. Otherwise it 
          //    designates the last element of the container.
          //    Returns a pointer to the new iterator if successful and
          //    NULL otherwise.

          virtual MIRContainerItem* moveToFirst();
          //    Moves the iterator to the first element of the container. 
          //    Returns a pointer to the first object if successful
          //    and NULL otherwise

          virtual MIRContainerItem* moveToLast();
          //    Moves the iterator to the last element of the container. 
          //    Returns a pointer to the last object if successful
          //    and NULL otherwise

          virtual MIRContainerItem* next(int nth=1);
          //    Moves the iterator to the nth element to the "right", starting 
          //    from the current position, and returns the corresponding object.
          //    If this object does not exist, because we reached the last
          //    element, the iterator points to the last element of the 
          //    container and NULL is returned.

          virtual MIRContainerItem* previous(int nth=1);
          //    Moves the iterator to the nth element to the "left", starting
          //    from the current position, and returns the corresponding object.
          //    If this object does not exist, because we reached the first
          //    element, the iterator points to the first element of the 
          //    container and NULL is returned.

          virtual MIRContainerItem* findFirst(const MIRContainerKey* key);
          //    Returns the first object with the given key in the container 
          //    and moves the iterator to the position of that object.
          //    If no such element can be found, NULL is returned and the 
          //    iterator is left unchanged.

          virtual MIRContainerItem* findLast(const MIRContainerKey* key);
          //    Returns the last object with the given key in the container
          //    and moves the iterator to the position of that object.
          //    If no such element can be found, NULL is returned and the 
          //    iterator is left unchanged.

          virtual MIRContainerItem* findNext(const MIRContainerKey* key);
          //    This method should only be called if you successfully called 
          //    one of the "find" operations on this iterator before.
          //
          //    Returns the next element with the given key starting from the 
          //    current position of the iterator and moves the iterator to the 
          //    position of that object.
          //    If no such element can be found, NULL is returned and the 
          //    iterator points to the last element with the given key.

          virtual MIRContainerItem* findPrevious(const MIRContainerKey* key);
          //    This method should only be called if you successfully called 
          //    one of the "find" operations on this iterator before.
          //
          //    Returns the previous element with the given key starting from the 
          //    current position of the iterator and moves the iterator to the 
          //    position of that object.
          //    If no such element can be found, NULL is returned and the 
          //    iterator points to the first element with the given key.


      protected:

          const MIRRBTreeNode* node;
    };


  protected:

      MIRRBTreeNode* _root;
      MIRRBTreeNode* _first;

  private:

      MIRRBTreeNode* _duplicateNode(MIRRBTreeNode* node, MIRRBTreeNode* old_null_node, MIRRBTreeNode* new_null_node);
      void _deleteNode(MIRRBTreeNode* node, MIRRBTreeNode* null_node);
      MIRRBTreeNode* _rotate(MIRRBTreeNode* gp, MIRRBTreeNode* p, MIRRBTreeNode* n);
      void _addFixup(MIRRBTreeNode* n, MIRRBTreeNode* p);
      void _removeFixup(MIRRBTreeNode* head, MIRRBTreeNode* n);
      void _dumpNode(const MIRString& prefix, const MIRRBTreeNode* node);
};

typedef MIRRBTree::Iterator MIRRBTreeIterator;

class MIR_STD_API MIRRBTreeNode
{

      friend class MIRRBTree;
      friend class MIRRBTree::Iterator;

  public:
      MIRContainerItem* getData() const;
      //    Returns a pointer to the object stored in this tree node

  private:
      MIRRBTreeNode();
      MIRRBTreeNode(MIRRBTreeNode* _left, MIRRBTreeNode* _right, MIRContainerItem* _data);
      MIRRBTreeNode(MIRRBTreeNode* _left, MIRRBTreeNode* _right, MIRRBTreeNode* _parent, MIRContainerItem* _data);
      static void* operator new(size_t size);
      static void operator delete(void* ptr);
#ifdef WIN32
      static void* operator new(size_t size, const char* file, int line);
      static void* operator new(size_t size, int block_type, const char* file, int line);
#if _MSC_VER >= 1200
      static void operator delete(void* ptr, const char* file, int line);
      static void operator delete(void* ptr, int block_type, const char* file, int line);
#endif
#endif

      MIRRBTreeNode* left;
      MIRRBTreeNode* right;
      MIRRBTreeNode* parent;
      MIRContainerItem* data;
};



/*
 * Inline methods
 */

#ifdef WIN32
inline void* MIRRBTree::operator new(size_t size, const char* file, int line)
{
    return operator new(size,_NORMAL_BLOCK,file,line);
}

#if _MSC_VER >= 1200
inline void MIRRBTree::operator delete(void* ptr, int block_type, const char* file, int line)
{
    operator delete(ptr);
}

inline void MIRRBTree::operator delete(void* ptr, const char* file, int line)
{
    operator delete(ptr);
}
#endif
#endif

inline MIRIterator* MIRRBTree::newIterator(MIRBoolean first) const
{
    return _new_dbg MIRRBTree::Iterator(*this,first);
}

inline void MIRRBTree::deleteIterator(MIRIterator* iterator) const
{
    delete iterator;
}

#ifdef WIN32
inline void* MIRRBTree::Iterator::operator new(size_t size, const char* file, int line)
{
    return operator new(size,_NORMAL_BLOCK,file,line);
}

#if _MSC_VER >= 1200
inline void MIRRBTree::Iterator::operator delete(void* ptr, int block_type, const char* file, int line)
{
    operator delete(ptr);
}

inline void MIRRBTree::Iterator::operator delete(void* ptr, const char* file, int line)
{
    operator delete(ptr);
}
#endif
#endif

inline MIRRBTree::Iterator::~Iterator()
{
}

inline MIRRBTreeNode::MIRRBTreeNode()
{
}

inline MIRRBTreeNode::MIRRBTreeNode(MIRRBTreeNode* _left, MIRRBTreeNode* _right, MIRContainerItem* _data):
    left(_left),right(_right),data(_data)
{
}

inline MIRRBTreeNode::MIRRBTreeNode(MIRRBTreeNode* _left, MIRRBTreeNode* _right, MIRRBTreeNode* _parent, MIRContainerItem* _data):
    left(_left),right(_right),parent(_parent),data(_data)
{
}

#ifdef WIN32
inline void* MIRRBTreeNode::operator new(size_t size, const char* file, int line)
{
    return operator new(size,_NORMAL_BLOCK,file,line);
}

#if _MSC_VER >= 1200
inline void MIRRBTreeNode::operator delete(void* ptr, int block_type, const char* file, int line)
{
    operator delete(ptr);
}

inline void MIRRBTreeNode::operator delete(void* ptr, const char* file, int line)
{
    operator delete(ptr);
}
#endif
#endif

inline MIRContainerItem* MIRRBTreeNode::getData() const
{
    return data;
}

#endif /* end of file */
