//####################################################################
//##     Copyright  Meta Integration Technology, Inc. 1997-2002    ##
//##                      All Rights Reserved.                      ##
//####################################################################
//## Meta Integration(tm) Model Bridge (MIMB), Works (MIW),         ##
//## Repository (MIR), Repository SDK (MIRSDK), and Database SDK    ##
//## (MIDBSDK) are trademarks of Meta Integration Technology, Inc.  ##
//## Other product and company names (or logos) mentioned herein    ##
//## may be the trademarks of their respective owners.              ##
//####################################################################
//##                 http://www.metaintegration.com                 ##
//####################################################################


#ifndef MIRSimpleHashtable_h
#define MIRSimpleHashtable_h 1

#include "MIRHashtable.h"

class MIR_STD_API MIRSimpleHashtable: public MIRHashtable
{
public:

    class Iterator;
    friend class Iterator;


    //## Construction

    MIRSimpleHashtable(MIRBoolean _addReplace=false, int _size=DEFAULT_SIZE);
    //    Creates an empty hashtable. Size is the size of the array to 
    //    allocate when the first hashtable element is added.


    //## Destruction

    virtual ~MIRSimpleHashtable();
    //    Releases the resources used by this hashtable. Note that this 
    //    operation does not release the memory used by the objects 
    //    that have been put in the hashtable.


    //## General functions

    virtual MIRBoolean empty();
    //    Removes all the elements that were in the hashtable.
    //    Returns true if successful and false otherwise.


    //## Creating and deleting an iterator for this container

    virtual MIRIterator* newIterator(MIRBoolean first=true) const;
    //    Allocates an iterator operating on this container. Note 
    //    that an iterator is specific to the container type that 
    //    allocated it.
    //    If first is true then the iterator is initialised so that it 
    //    points to the first element in the container. Otherwise it 
    //    designates the last element of the container.
    //    Returns a pointer to the new iterator if successful and
    //    NULL otherwise.

    //## Iterator based methods for adding and removing elements
    //## (Container independant)

    virtual MIRContainerItem* findFirst(const MIRContainerKey* key, MIRIterator* iterator=NULL) const;
    //    Looks for the first object with the given key.
    //    Returns a pointer to this object if found and sets the 
    //    iterator to the position of that object. Repeatedly calling 
    //    this iterator nextElement(key) will allow you to enumerate 
    //    all this container's objects that have the same key.
    //    Returns NULL and leaves the iterator unchanged if no object
    //    with this key was found.

    virtual MIRContainerItem* findLast(const MIRContainerKey* key, MIRIterator* iterator=NULL) const;
    //    Looks for the last object with the given key.
    //    Returns a pointer to this object if found and sets the 
    //    iterator to the position of that object. Repeatedly calling 
    //    this iterator previousElement(key) will allow you to enumerate 
    //    all this container's objects that have the same key.
    //    Returns NULL and leaves the iterator unchanged if no object
    //    with this key was found.

    virtual MIRContainerItem* remove(MIRIterator* iterator);
    //    Removes the element designated by iterator.
    //    Returns a pointer to that object if successful. The iterator 
    //    then points to the element replacing the removed element or 
    //    to the last element in the container if the removed element 
    //    was the last one.
    //    If the operation fails it returns NULL and the iterator is 
    //    left unchanged.


    class MIR_STD_API Iterator: public MIRHashtable::Iterator
    {
    public:

        friend class MIRSimpleHashtable;

        //## Construction

        Iterator();
        //    Creates a new iterator not associated to a container.

        Iterator(const MIRHashtable& htable, MIRBoolean first=true);
        //    Creates a new iterator pointing to the first element 
        //    in the hash table if first is true or the last otherwise.

        //## Destruction

        virtual ~Iterator();
        //    Releases the resources used by this iterator.


        //## General functions

        virtual MIRContainerItem* operator*() const;
        virtual MIRContainerItem* get() const;
        //    Returns a pointer to the object for the current position
        //    of the iterator or NULL if the iterator is invalid.

        virtual void* getKey() const;
        //    Returns a pointer to the key for the current position
        //    of the iterator or NULL if the iterator is invalid.
		//    Currently either a long* or MIRString*

        //## Methods for moving iterators

        virtual MIRContainerItem* moveTo(MIRHashtable& htable, MIRBoolean first=true);
        //    Moves the iterator to the specified container.
        //    If first is true then the iterator is initialised so that it 
        //    points to the first element in the container. Otherwise it 
        //    designates the last element of the container.
        //    Returns a pointer to the new iterator if successful and
        //    NULL otherwise.

        virtual MIRContainerItem* moveToFirst();
        //    Moves the iterator to the first element of the container. 
        //    Returns a pointer to the first object if successful
        //    and NULL otherwise

        virtual MIRContainerItem* moveToLast();
        //    Moves the iterator to the last element of the container. 
        //    Returns a pointer to the last object if successful
        //    and NULL otherwise

        virtual MIRContainerItem* next(int nth=1);
        //    Moves the iterator to the nth element to the "right", starting 
        //    from the current position, and returns the corresponding object.
        //    If this object does not exist, because we reached the last
        //    element, the iterator points to the last element of the 
        //    container and NULL is returned.

        virtual MIRContainerItem* previous(int nth=1);
        //    Moves the iterator to the nth element to the "left", starting
        //    from the current position, and returns the corresponding object.
        //    If this object does not exist, because we reached the first
        //    element, the iterator points to the first element of the 
        //    container and NULL is returned.

        virtual MIRContainerItem* findFirst(const MIRContainerKey* key);
        //    Returns the first object with the given key in the container 
        //    and moves the iterator to the position of that object.
        //    If no such element can be found, NULL is returned and the 
        //    iterator is left unchanged.

        virtual MIRContainerItem* findLast(const MIRContainerKey* key);
        //    Returns the last object with the given key in the container
        //    and moves the iterator to the position of that object.
        //    If no such element can be found, NULL is returned and the 
        //    iterator is left unchanged.

        virtual MIRContainerItem* findNext(const MIRContainerKey* key);
        //    This method should only be called if you successfully called 
        //    one of the "find" operations on this iterator before.
        //
        //    Returns the next element with the given key starting from the 
        //    current position of the iterator and moves the iterator to the 
        //    position of that object.
        //    If no such element can be found, NULL is returned and the 
        //    iterator points to the last element with the given key.

        virtual MIRContainerItem* findPrevious(const MIRContainerKey* key);
        //    This method should only be called if you successfully called 
        //    one of the "find" operations on this iterator before.
        //
        //    Returns the previous element with the given key starting from the 
        //    current position of the iterator and moves the iterator to the 
        //    position of that object.
        //    If no such element can be found, NULL is returned and the 
        //    iterator points to the first element with the given key.
    };
protected:
    MIRBoolean addReplace;
};

typedef MIRSimpleHashtable::Iterator MIRSimpleHashtableIterator;

/*
 * Inline methods
 */

inline MIRIterator* MIRSimpleHashtable::newIterator(MIRBoolean first) const
{
    return _new_dbg MIRSimpleHashtable::Iterator(*this,first);
}

class MIR_STD_API MIRSimpleStringHashtable: public MIRSimpleHashtable
{
public:
    MIRSimpleStringHashtable(MIRBoolean _ignoreCase=false, MIRBoolean _addReplace=false, int _size=DEFAULT_SIZE);

    MIRBoolean add(const MIRTString& key, void* value);
protected:
    MIRBoolean ignoreCase;
	//## Methods we have to extend to make the hashtable functional
	int compare2Object(const MIRContainerItem* ptr1, const MIRContainerItem* ptr2) const;
	virtual int compare2Key(const MIRContainerItem* ptr, const MIRContainerKey* key) const;
	virtual unsigned int hashObject(const MIRContainerItem* ptr) const;
	virtual unsigned int hashKey(const MIRContainerKey* key) const;
};

class MIR_STD_API MIRSimpleLongHashtable: public MIRSimpleHashtable
{
public:    
    MIRSimpleLongHashtable(MIRBoolean _addReplace=false, int _size=DEFAULT_SIZE);

	MIRBoolean add(const long key, void* value);
protected:    
	//## Methods we have to extend to make the hashtable functional
	int compare2Object(const MIRContainerItem* ptr1, const MIRContainerItem* ptr2) const;
	virtual int compare2Key(const MIRContainerItem* ptr, const MIRContainerKey* key) const;
	virtual unsigned int hashObject(const MIRContainerItem* ptr) const;
	virtual unsigned int hashKey(const MIRContainerKey* key) const;
};

#endif
