//####################################################################
//##     Copyright  Meta Integration Technology, Inc. 1997-2002    ##
//##                      All Rights Reserved.                      ##
//####################################################################
//## Meta Integration(tm) Model Bridge (MIMB), Works (MIW),         ##
//## Repository (MIR), Repository SDK (MIRSDK), and Database SDK    ##
//## (MIDBSDK) are trademarks of Meta Integration Technology, Inc.  ##
//## Other product and company names (or logos) mentioned herein    ##
//## may be the trademarks of their respective owners.              ##
//####################################################################
//##                 http://www.metaintegration.com                 ##
//####################################################################



#ifndef MIRString_h
#define MIRString_h 1

#include <ctype.h>
#include <string.h>
#if defined(MPRAS)
#include <strings.h>
#endif
#include <wchar.h>

#include "MIRStdDefs.h"
#include "MIRIO.h"

#ifdef WIN32
#    define TOASCII    __toascii
#    define TOLOWER    tolower
#    define TOUPPER    toupper
#else
#    define TOASCII    toascii
#    define TOLOWER    tolower
#    define TOUPPER    toupper 
#endif

/*
 * This file provides an easy, safe and efficient class to manipulate 
 * strings. It contains three classes:
 *  - MIRString: This is the class you will use to manipulate strings.
 *
 *  - MIRTString: This is another string class which is used to 
 *    manipulate transient strings. It's more efficient in some 
 *    contexts but it is also more dangerous. You can safely ignore 
 *    this class when writing code.
 *
 *  - MIRGString: This is the parent class for both MIRString and 
 *    MIRTString. Its sole purpose is to share code between MIRString
 *    and MIRTString.
 *
 *                          *** WARNING ***
 * MIRString and MIRTString behave very differently. Changing the 
 * type of a variable from MIRString to MIRTString is very likely to 
 * break your program. For all practical purposes though you can use 
 * a MIRString, wherever you see a MIRTString, i.e. any function that 
 * takes a MIRTString as a parameter will also accept a MIRString 
 * and the result of a function that returns a MIRTString can be put 
 * in a MIRString.
 *
 * MIRString is designed so that most operations can be performed 
 * naturally. Here are a few examples:
 *
 *  MIRString a,b;
 *  a="xxx";        // a == "xxx"
 *  b=a;            // b == "xxx"
 *  a+="-"+b;       // a == "xxx-xxx"
 *  if (a=="xxx-xxx")
 *      cout << "The result is :" << a << endl;
 *  else
 *      cout << "An error occurred :" << a << endl;
 *
 *
 * Notes:
 *  - It's better never to explicitly construct a MIRString from a 'char*'. 
 *    This is unnecessary, makes the code less readable and, in some 
 *    cases, may lower the performance. Just write:
 *          extern void print(MIRString& str);
 *          ...
 *          print("See.. no constructor");
 *
 *  - There is one place where you must explicitly cast a MIRString to 
 *    'const char*': when it is a parameter of printf, sprintf or  
 *    anywhere in the variable part of a function that takes a variable 
 *    number of parameters. Thus you must write:
 *          MIRString str("a string");
 *          char buf[1024];
 *          sprintf(buf, "Don't forget the cast: %s\n", (const char*)str);
 *
 *  - When speaking of string length the trailing '\0' is not counted. Thus
 *    the string "a" contains one character and can fit in a string of 
 *    capacity 1. The character buffer, though, will of course be of size 2.
 */

const unsigned DEFAULT_STRING_EXTRA = 0;
const unsigned DEFAULT_TSTRING_EXTRA = 0;

class MIR_STD_API MIRString;
class MIR_STD_API MIRTString;

/*
 * MIRGString class. This is an abstract class used as a parent for both
 * MIRString and MIRTString.
 */
class MIR_STD_API MIRGString
{

      friend class MIRString;
      friend class MIRTString;

  public:

    //## Print operation
      friend ostream MIR_STD_API &operator<<(ostream &out, const MIRGString &s);

    //## Comparison

      bool operator==(const MIRGString& s) const;
      bool operator!=(const MIRGString& s) const;
      bool operator<(const MIRGString& s) const;
      bool operator>(const MIRGString& s) const;
      bool operator<=(const MIRGString& s) const;
      bool operator>=(const MIRGString& s) const;
      //    These operators compare two MIRGStrings and return 
      //    true or false as appropriate.

      bool operator==(const char* cs) const;
      bool operator!=(const char* cs) const;
      bool operator<(const char* cs) const;
      bool operator>(const char* cs) const;  
      bool operator<=(const char* cs) const;
      bool operator>=(const char* cs) const;
      //    These operators compare MIRGString to a char* and return 
      //    true or false as appropriate.

      friend bool MIR_STD_API operator==(const char* cs, const MIRGString& s);
      friend bool MIR_STD_API operator!=(const char* cs, const MIRGString& s);
      friend bool MIR_STD_API operator<(const char* cs, const MIRGString& s);
      friend bool MIR_STD_API operator>(const char* cs, const MIRGString& s);
      friend bool MIR_STD_API operator<=(const char* cs, const MIRGString& s);
      friend bool MIR_STD_API operator>=(const char* cs, const MIRGString& s);
      //    These operators compare a char* to MIRGString and return 
      //    true or false as appropriate.

      int compare(const MIRGString& s) const;
      int compare(const char* cs) const;
      //    Compares this MIRGString to the string given as a parameter
      //    and returns:
      //        <0 if   this <  s/cs
      //         0 if   this == s/cs
      //        >0 if   this >  s/cs

      int compareNoCase(const MIRGString& s) const;
      int compareNoCase(const char* cs) const;
      //    Compares this MIRGString to the string given as a parameter
      //    ignoring the case and returns:
      //        <0 if   this <  s/cs
      //         0 if   this == s/cs
      //        >0 if   this >  s/cs

      int compareSubstring(const MIRGString& s) const;
      int compareSubstring(const char* cs) const;
      //    Check if this MIRGString starts with the specified characters
      //    and returns:
      //        <0 if   this (up to strlen(s/cs)) <  s/cs
      //         0 if   this (up to strlen(s/cs)) == s/cs
      //        >0 if   this (up to strlen(s/cs)) >  s/cs

      int compareSubstringNoCase(const MIRGString& s) const;
      int compareSubstringNoCase(const char* cs) const;
      //    Check if this MIRGString starts with the specified characters,
      //    ignoring the case, and returns:
      //        <0 if   this (up to strlen(s/cs)) <  s/cs
      //         0 if   this (up to strlen(s/cs)) == s/cs
      //        >0 if   this (up to strlen(s/cs)) >  s/cs

      unsigned int hashCode() const;
      //    Computes a hash code for this string.


    //## Concatenation

      MIRGString& add(const MIRGString* first, ...);
      MIRGString& add(const char* first, ...);
      //    Concatenates the list of strings given in parameter to the 
      //    end of this string and returns a reference to this string.
      //    The string list must be terminated by a NULL.

      MIRGString& add(const MIRGString* const * strings);
      MIRGString& add(const char* const * strings);
      //    Concatenates the strings in the array to the end of this 
      //    string and returns a reference to this string.
      //    The string array must be terminated by a NULL.

      MIRGString& append(const char* string, int count);
      //    Appends the first count characters of string to the end of this 
      //    string and returns a reference to this string.

      MIRGString& append(char c);
      //    Appends the specified character to the end of this 
      //    string and returns a reference to this string.
      
      MIRGString& append(char c, int count);
      //    Appends count characters c to the end of this 
      //    string and returns a reference to this string.


    //## General methods

      unsigned length() const;
      //    Returns the length of the string, i.e. the number of character 
      //    it contains not counting the trailing '\0'.

      bool isEmpty() const;
      //    Returns true if the string is empty, i.e. if its length is 0, 
      //    and false otherwise.

      void empty();
      //    Empties the string.

      void truncate(unsigned new_length);
      //    Truncates the string to the specified length. If the specified 
      //    length is longer than the current length then the length is 
      //    unchanged.

      unsigned resize(unsigned new_capacity);
      //    Resizes the character buffer so that it can hold string of 
      //    length up to 'new_capacity' characters. If new_capacity is
      //    less than the current string length then the character buffer 
      //    is shortened to its minimum size.
      //    The returned value is the actual capacity of the string after
      //    this call.

      unsigned capacity() const;
      //    Returns the length of the longest string the character buffer 
      //    can hold before it has to be reallocated.
      //    Returns 0 for wrapper strings (see the wrap function below).

      void printOn(int level=0) const;
      //    Prints debugging information for this MIRGString
      //    object to the standard error stream.


    //## Conversion

      operator const char*() const;
      const char* chars() const;
      const unsigned char* unsignedChars() const;
      //    Return a pointer to the '\0' terminated character buffer.


    //## Character manipulation

      void setAt(unsigned i, const char c);
      //    Sets the character at the 0-based position i if i < length.
      //    Does nothing otherwise.

      char getAt(unsigned i) const;
      //    Returns the character at the 0-based position i if i < length.
      //    Returns the trailing '\0' otherwise.


    //## Character search

      int find(const char c) const;
      int find(const char c, unsigned int pos) const;
      int reverseFind(const char c) const;
      //    Respectively find the first and last character c
      //    in the string and return its 0-based position.
      //    Returns -1 if not found.

      int find(const char* cs) const;
      int find(const char* cs, unsigned int pos) const;
      //    Finds the first occurence of the substring cs in this
      //    string and returns the 0-based position of its first
      //    character.
      //    Returns -1 if not found.


    //## Substring extraction

      MIRTString mid(unsigned int first, unsigned int count) const;
      //    Returns a string containing the count characters starting 
      //    from the 0-based position first.
      //    If first > length() or count == 0 mid returns an empty string.
      //    Note:
	  //     - this string is not modified
      //     - count also specifies the size of the new buffer so if 
      //       it is greater than length()-first+1 more space than 
      //       necessary may be allocated.

      MIRTString mid(unsigned int first) const;
      //    Returns a string containing the characters starting from 
      //    the 0-based position first to the end of the string.
      //    If first > length() mid returns an empty string.
      //    Note:
	  //     - this string is not modified

      MIRTString left(unsigned int count) const;
      //    Returns a string containing the count characters starting from 
      //    the beginning of the string.
      //    If count == 0 left returns an empty string.
      //    Note:
	  //     - this string is not modified
      //     - count also specifies the size of the new buffer so if 
      //       it is greater than length() more space than necessary 
      //       may be allocated.

      MIRTString right(unsigned int count) const;
      //    Returns a string containing the count characters starting from 
      //    the end of the string.
      //    If count == 0 right returns an empty string.
      //    Note:
	  //     - this string is not modified
      //     - count also specifies the size of the new buffer so if 
      //       it is greater than length() more space than necessary 
      //       may be allocated.


    //## Other transformations

      MIRGString& toLower();
      MIRGString& toUpper();
      //    Converts this string to all upper/lower case characters
	  //    and returns a reference to this string.

      MIRGString& toAscii();
      //    Converts this string to ascii, i.e. it converts
      //    each character to a 7 bit ascii character
	  //    and returns a reference to this string.

      MIRGString& reverse();
      //    Reverses the order of this string's characters
	  //    and returns a reference to this string.

      MIRTString escapeChars() const;
      //    Returns a string in which the following characters are escaped:
      //    '\'   ->  '\' '\'
      //    '"'   ->  '\' '"'
      //    '\n'  ->  '\' 'n'
      //    '\r'  ->  '\' 'r'

      MIRTString escapeChar(char ch, char esc, bool escEsc = true) const;
      //    Returns a string in which [ch] characters are escaped like [esc][ch].
      //    if [escEsc] == true, all [esc] characters are escaped as [esc][esc].
      //    Following expression is always true
      //    str.escapeChar(';','\\')
      //       .escapeChar(',','\\', false)
      //       .escapeChar('.','\\', false)
      //       .unescapeChar('\\') == str
      // IMPORTANT NOTE: to escape char in the string for a first time escEsc must
      //    be set to true. All subsequent calls to escape other chars with the
      //    same escape characters must be done with escEsc set to false.

      MIRTString unescapeChar(char esc) const;
      //    Does opposite to escapeChar(char ch, char esc) action by removing
      //    [esc] characters back.

      MIRTString trim() const;
      //    Return a string containing this string trimmed from all leading 
      //    and trailing control characters, i.e. all characters <=32.
      //    This can for instance be used to trim leading and trailing spaces 
      //    and tabs.

      MIRTString trimSpaces() const;
      //    Same as trim but only removes leading and trailing spaces.



    //## Unsafe operations

      /*
       * The following functions are provided for the rare cases where they
       * are needed to achieve optimal performance.
       * Extreme care must be taken when using them though since they make 
       * it possible to directly manipulate the string's buffer thus 
       * potentially making its internal state inconsistent.
       */

      void wrap(const MIRGString& s);
      //    Initialises this string so that it refers to the same buffer
      //    as s. You must not change s as this would make this string 
      //    inconsistent. Any subsequent change to this string will cause 
      //    the allocation of a new buffer to store the modified string.

      void wrap(const char* cs);
      //    Initialises this string so that it uses cs as its buffer.
      //    You must not change cs as this would make this string 
      //    inconsistent. Any subsequent change to this string will cause the 
      //    allocation of a new buffer to store the modified string. You are 
      //    responsible for freeing cs when appropriate.

      bool unwrap(unsigned extra);
      //    If you called wrap before, unwrap will allocate a new buffer to 
      //    store the buffer and return true. This string will thereafter be 
      //    self-contained as opposed to being a "wrapper string".
      //    If this string was not a "wrapper string" unwrap does nothing and 
      //    returns false.

      char* lock();
      //    Locks the string and returns a pointer to its buffer for 
      //    modification. Make sure you do not read/write past the end of the 
      //    buffer. To get the buffer capacity, even while the string is 
      //    locked, call capacity. While the string is locked it is in an 
      //    inconsistent state and the behavior of its methods is 
      //    unpredictable (even for length or resize). Once you have finished 
      //    modifying the buffer you should call unlock.

      bool isLocked() const;
      //    Returns true if the string is locked and false otherwise.

      bool unlock(bool full=false);
      //    Unlocks a previously locked string and returns true.
      //    Returns false if the string was not locked.
      //    If full is true then the string is completely unlocked regardless 
      //    of the number of lock calls made before. Otherwise there must be 
      //    one unlock per lock call.


    //## Miscellaneous friends

      friend MIRTString MIR_STD_API operator+(const MIRString& s1, const MIRString& s2);
      friend MIRTString MIR_STD_API operator+(const MIRString& s1, const MIRTString& s2);
      friend MIRTString MIR_STD_API operator+(const MIRTString& s1, const MIRString& s2);
      friend MIRTString MIR_STD_API operator+(const MIRString& s, const char* cs);
      friend MIRTString MIR_STD_API operator+(const char* cs, const MIRString& s);
      friend MIRTString MIR_STD_API operator+(const MIRString& s, const int cs);
      friend MIRTString MIR_STD_API operator+(const int cs, const MIRString& s);
      friend MIRTString MIR_STD_API operator+(const MIRString& s, const long cs);
      friend MIRTString MIR_STD_API operator+(const long cs, const MIRString& s);

      friend MIRTString MIR_STD_API operator+(const MIRTString& s1, const MIRTString& s2);
      friend MIRTString MIR_STD_API operator+(const MIRTString& s, const char* cs);
      friend MIRTString MIR_STD_API operator+(const char* cs, const MIRTString& s);
      friend MIRTString MIR_STD_API operator+(const MIRTString& s, const int cs);
      friend MIRTString MIR_STD_API operator+(const int cs, const MIRTString& s);
      friend MIRTString MIR_STD_API operator+(const MIRTString& s, const long cs);
      friend MIRTString MIR_STD_API operator+(const long cs, const MIRTString& s);

      //    These operators will be declared in MIRString or MIRTString but have to
      //    be redeclared here because they need to be MIRGString friends.

  protected:

      char* p;               // character string
      unsigned len;          // length of string, excluding null character
      unsigned alloc;        // amount of storage allocated

      static char* emptyPtr; // an empty buffer
};



/*
 * MIRString class.
 */
class MIR_STD_API MIRString: public MIRGString
{

      friend class MIRTString;

  private:

      void init(const char* src);
      void init(const char* src, int count);

  public:

    //## Construction

      //    In all the MIRString constructors extra specifies the number of
      //    extra bytes to allocate for the buffer.

      MIRString();
      MIRString(unsigned capacity);
      //    Creates an empty string. If specified, capacity denotes the 
      //    number of characters that the string can hold before we have 
      //    to resize the buffer.

      MIRString(const char* src);
      MIRString(const unsigned char* src);
      MIRString(const char* src, int count);
      MIRString(const unsigned char* src, int count);
      MIRString(const MIRString& src);
      MIRString(const MIRString& src, int count);
      //    Initialises this string with the contents of src. This implies 
      //    allocating a new buffer and copying the contents of src in this 
      //    new buffer. If count is specified then only the first count 
      //    characters of the string are taken into account. If count is 
      //    negative then the behavior is the same as if count was not 
      //    specified, if count is 0 then the result is an empty string.

      MIRString(char c, unsigned l);
      //    Initialises a string containing l times the character c.

      MIRString(const MIRTString& src);
      //    Initialises this string with the contents of src. This implies
      //    taking over the src buffer and making src a wrapper string for 
      //    this buffer.
      //    See MIRTString for more information.

      static void* operator new(size_t size);
      static void operator delete(void* ptr);
#ifdef WIN32
      static void* operator new(size_t size, const char* file, int line);
      static void* operator new(size_t size, int block_type, const char* file, int line);
#if _MSC_VER >= 1200
      static void operator delete(void* ptr, const char* file, int line);
      static void operator delete(void* ptr, int block_type, const char* file, int line);
#endif
#endif
      //    Used to allocate/deallocate MIRString objects (and subclasses).

      void operator=(const MIRString& right);
      void operator=(const char* right);
      //    Copies the contents of right to this string. This implies 
      //    allocating/reallocating a new buffer and copying the contents of 
      //    right in this new buffer.

      void operator=(const MIRTString& right);
      //    Assigns right to this string. This implies taking over the right 
      //    buffer and making right a wrapper string for this buffer.
      //    See MIRTString for more information.

      void MIRString::operator=(const int right);
      void MIRString::operator=(const long right);

    //## Destruction

      ~MIRString();
      //    Releases the resources used by this object.


    //## Concatenation

      const MIRString& operator+=(const MIRString& right);
      const MIRString& operator+=(const MIRTString& right);
      const MIRString& operator+=(const char* right);
      const MIRString& operator+=(const int right);
      const MIRString& operator+=(const long right);
      //    Concatenates the string given as a parameter to this 
      //    MIRString object and returns a reference to this object.
      //    Note:
      //     - s1+=s2 is much more efficient than s1=s1+s2

      friend MIRTString MIR_STD_API operator+(const MIRString& s1, const MIRString& s2);
      friend MIRTString MIR_STD_API operator+(const MIRString& s1, const MIRTString& s2);
      friend MIRTString MIR_STD_API operator+(const MIRString& s, const char* cs);
      friend MIRTString MIR_STD_API operator+(const char* cs, const MIRString& s);
      friend MIRTString MIR_STD_API operator+(const MIRString& s, const int cs);
      friend MIRTString MIR_STD_API operator+(const int cs, const MIRString& s);
      friend MIRTString MIR_STD_API operator+(const MIRString& s, const long cs);
      friend MIRTString MIR_STD_API operator+(const long cs, const MIRString& s);
      //    Concatenates the string given as a parameter to this 
      //    MIRString object and returns a new string containing the result.


    //## Unsafe operations

      /*
       * The following functions are provided for the rare cases where they
       * are needed to achieve optimal performance.
       * Extreme care must be taken when using them though since they make 
       * it possible to directly manipulate the string's buffer thus 
       * potentially making its internal state inconsistent.
       */

      MIRString(bool wrap, const MIRString& src);
      //    Initialises this string so that it refers to the same buffer
      //    as src if wrap is true. You must not change src as it would 
      //    make this string inconsistent. Any subsequent change to this 
      //    string will cause the allocation of a new buffer to store the 
      //    modified string.
      //    Equivalent to MIRString::MIRString(const MIRString& src) if 
      //    wrap is false.

      MIRString(bool wrap, const char* cs);
      //    Initialises this string so that it uses cs as its buffer if wrap 
      //    is true. You must not change cs as this would make this string 
      //    inconsistent. Any subsequent change to this string will cause the 
      //    allocation of a new buffer to store the modified string. You are 
      //    responsible for freeing cs when appropriate.
      //    Equivalent to MIRString::MIRString(const char* cs) if 
      //    wrap is false.

#     define MIRWString(str) MIRString(true,str)
      //    A shortcut for the wrapper constructors


    //## Miscellaneous friends

      friend MIRTString MIR_STD_API operator+(const MIRTString& s1, const MIRString& s2);
      //    This operator is declared in MIRTString but has to be redeclared 
      //    here because it needs to be a MIRString friend.
};

/*
 * MITString class.
 *
 * This class is used as an optimised way to store transient string results. 
 * For instance 'operator+' declares a local variable to store the result of 
 * the concatenation and returns this variable. When returning this variable 
 * the copy constructor is automatically called.
 * Using a MIRString for this local variable would cause the copy constructor
 * MIRString::MIRString(MIRString) to be called which means reallocating 
 * a new buffer and copying the local variable's buffer. Right after that 
 * the local variable's destructor is called anyway.
 * Using a MIRTString for the local variable MIRString::MIRString(MIRTString) 
 * will be called instead and this copy constructor will not duplicate the 
 * MIRTString buffer because it knows that MIRTStrings are used only once.
 * Instead it takes over the MIRTString buffer, makes the MIRTString a wrapper
 * around this buffer. This also means the when the MIRTString destructor is 
 * called it's almost a no-operation. All in all there is no reallocation or
 * copying involved.
 *
 * All the operations involving a MIRTString assume that the MIRTString buffer
 * can be modified at will which leads to a very different behavior from a 
 * MIRString.
 */
class MIR_STD_API MIRTString: public MIRGString
{

      friend class MIRString;
	  friend class MIRGString;

  private:

      void init(const char* src);
      void init(const char* src, int count);

  public:

    //## Construction

      MIRTString();
      MIRTString(unsigned capacity);
      //    Creates an empty string. If specified, capacity denotes the 
      //    number of characters that the string can hold before we have 
      //    to resize the buffer.

      MIRTString(const MIRString& src);
      MIRTString(const MIRTString& src);
      MIRTString(const char* src);
      MIRTString(const unsigned char* src);
      MIRTString(const char* src, int count);
      MIRTString(const unsigned char* src, int count);
      //    Initialises this transient string so that it refers to the 
      //    same buffer as src. You must not change src as it would make 
      //    this string inconsistent. Any subsequent change to this string 
      //    will cause the allocation of a new buffer to store the modified 
      //    string. If count is specified then only the first count 
      //    characters of the string are taken into account. If count is 
      //    negative then the behavior is the same as if count was not 
      //    specified, if count is 0 then the result is an empty string.

      void operator=(const MIRString& right);
      void operator=(const MIRTString& right);
      void operator=(const char* right);
      //    Initialises this transient string so that it refers to the 
      //    same buffer as right. You must not change right as it would make 
      //    this string inconsistent. Any subsequent change to this string 
      //    will cause the allocation of a new buffer to store the modified 
      //    string.
      void operator=(const int right);
      void operator=(const long right);

    //## Destruction

      ~MIRTString();
      //    Releases the resources used by this object.


    //## Concatenation

      MIRTString& operator+=(const MIRGString& right);
      MIRTString& operator+=(const char* right);
      MIRTString& operator+=(const int right);
      MIRTString& operator+=(const long right);
      //    Concatenates the string given as a parameter to this 
      //    MIRTString object and returns a reference to this object.
      //    Note:
      //     - s1+=s2 is much more efficient than s1=s1+s2

      friend MIRTString MIR_STD_API operator+(const MIRTString& s1, const MIRString& s2);
      friend MIRTString MIR_STD_API operator+(const MIRTString& s1, const MIRTString& s2);
      friend MIRTString MIR_STD_API operator+(const MIRTString& s, const char* cs);
      friend MIRTString MIR_STD_API operator+(const MIRTString& s, const int cs);
      friend MIRTString MIR_STD_API operator+(const MIRTString& s, const long cs);
      //    Appends the second string given in paramter to the MIRTString 
      //    object and returns this MIRTString parameter. Note that this 
      //    modifies the left hand argument of this operator thus behaving 
      //    more like 'operator+='. 

      friend MIRTString MIR_STD_API operator+(const char* cs, const MIRTString& s);
      friend MIRTString MIR_STD_API operator+(const int cs, const MIRTString& s);
      friend MIRTString MIR_STD_API operator+(const long cs, const MIRTString& s);
      //    Concatenates the MIRTString paramter to the end of cs. The result 
      //    is stored in the MIRTString object which is then returned.


    //## Miscellaneous friends

      friend MIRTString MIR_STD_API operator+(const MIRString& s1, const MIRString& s2);
      friend MIRTString MIR_STD_API operator+(const MIRString& s1, const MIRTString& s2);
      friend MIRTString MIR_STD_API operator+(const MIRString& s, const char* cs);
      friend MIRTString MIR_STD_API operator+(const char* cs, const MIRString& s);
      friend MIRTString MIR_STD_API operator+(const MIRString& s, const int cs);
      friend MIRTString MIR_STD_API operator+(const int cs, const MIRString& s);
      friend MIRTString MIR_STD_API operator+(const MIRString& s, const long cs);
      friend MIRTString MIR_STD_API operator+(const long cs, const MIRString& s);
      //    These operators have been declared in MIRString but have to
      //    be redeclared here because they need to be MIRTString friends.
};


extern MIR_STD_API MIRString emptyString;
//    This is an empty string.


extern MIR_STD_CAPI unsigned int stringHashCode(const char* c);
//    Computes a hash code for the specified string.

#define stringCompare(a,b)      strcmp(a,b)
#define stringNCompare(a,b,c)   strncmp(a,b,c)

#ifdef WIN32
#define stringICompare(a,b)     _stricmp(a,b)
#define stringNICompare(a,b,c)  _strnicmp(a,b,c)
#else
#define stringICompare(a,b)     strcasecmp(a,b)
#define stringNICompare(a,b,c)  strncasecmp(a,b,c)
#endif

#if defined(WIN32)
#define VSNPRINTF(buffer,size,format,arguments) _vsnprintf(buffer,size,format,arguments)
#elif defined(linux)
#define VSNPRINTF(buffer,size,format,arguments) vsnprintf(buffer,size,format,arguments)
#else
#define VSNPRINTF(buffer,size,format,arguments) vsprintf(buffer,format,arguments)
#endif

inline MIRTString toString(int val)
{
    MIRTString tmp;
    tmp=val;
    return tmp;
}

inline MIRTString toString(long val)
{
    MIRTString tmp;
    tmp=val;
    return tmp;
}

class MIR_STD_API MIREncodingConverter
{
public:
    //## Construction

    MIREncodingConverter(const MIRTString& encodingName);
    //  Creates a converter for the given encoding name.
    //  If the specified encoding name is empty, the locale encoding will be used.

    //## Destruction

    ~MIREncodingConverter();
    //  Releases the resources used by this object.

    MIRBoolean isValid();
    //  Returns true if the object is a valid converter for encoding conversion, false otherwise.

    const char* getEncodingName();
    //  Returns the canonical name of the converter

    //## Miscellaneous friends

    friend MIRTString MIR_STD_API convertEncoding(const MIRTString& value, MIREncodingConverter* srcConverter, MIREncodingConverter* dstConverter, int& cnvCnt, int& errCnt);
    friend MIRBoolean MIR_STD_API equivalentConverters(MIREncodingConverter* srcConverter, MIREncodingConverter* dstConverter);
private:
    void* converter;        // the converter
    const char* name;       // the canonical name of the converter
    int maxCharLength;      // the maximum length, in bytes, of a character in that encoding
};

inline MIRBoolean MIREncodingConverter::isValid()
{
    return (converter!=NULL);
}

// Convert a string from the source encoding to the destination encoding
extern MIR_STD_API MIRTString convertEncoding(const MIRTString& value, MIREncodingConverter* srcConverter, MIREncodingConverter* dstConverter, int& cnvCnt, int& errCnt);

// Compare two converters and return true if they are equivalent, false otherwise
extern MIR_STD_API MIRBoolean equivalentConverters(MIREncodingConverter* srcConverter, MIREncodingConverter* dstConverter);

// Two helper functions to convert between utf8 and the locale encoding
// Very useful for file & directory handling as long as we don't have our own classes to handle them
extern MIR_STD_API MIRTString localeToUtf8(const char* string);
extern MIR_STD_API MIRTString utf8ToLocale(const char* string);

extern MIR_STD_API wchar_t* utf8ToUtf16(const char* string);
extern MIR_STD_API void freeUtf16(wchar_t* string);
extern MIR_STD_API char* utf16ToUtf8(const wchar_t* string);

/*
 * MIRGString Inline methods
 */

inline ostream &operator<<(ostream &out, const MIRGString &s)
{
    out << s.p;
    return out;;
}

//## MIRGString Comparison

inline bool MIRGString::operator==(const MIRGString& s) const
{
    return (len==s.len) && (stringCompare(p,s.p)==0);
}

inline bool MIRGString::operator!=(const MIRGString& s) const
{
    return !(*this==s);
}

inline bool MIRGString::operator<(const MIRGString& s) const
{
    return stringCompare(p, s.p) < 0;
}

inline bool MIRGString::operator>(const MIRGString& s) const
{
    return stringCompare(p, s.p) > 0;
}

inline bool MIRGString::operator<=(const MIRGString& s) const
{
    return stringCompare(p, s.p) <= 0;
}

inline bool MIRGString::operator>=(const MIRGString& s) const
{
    return stringCompare(p, s.p) >= 0;
}

inline bool MIRGString::operator==(const char* cs) const
{
    return stringCompare(p,cs) == 0;
}

inline bool MIRGString::operator!=(const char* cs) const
{
    return stringCompare(p,cs) != 0;
}

inline bool MIRGString::operator<(const char* cs) const
{
    return stringCompare(p,cs) < 0;
}

inline bool MIRGString::operator>(const char* cs) const
{
    return stringCompare(p,cs) > 0;
}

inline bool MIRGString::operator<=(const char* cs) const
{
    return stringCompare(p,cs) <= 0;
}

inline bool MIRGString::operator>=(const char* cs) const
{
    return stringCompare(p,cs) >= 0;
}

inline bool operator==(const char* cs, const MIRGString& s)
{
    return stringCompare(cs, s.p) == 0;
}

inline bool operator!=(const char* cs, const MIRGString& s)
{
    return stringCompare(cs, s.p) != 0;
}

inline bool operator<(const char* cs, const MIRGString& s)
{
    return stringCompare(cs, s.p) < 0;
}

inline bool operator>(const char* cs, const MIRGString& s)
{
    return stringCompare(cs, s.p) > 0;
}

inline bool operator<=(const char* cs, const MIRGString& s)
{
    return stringCompare(cs, s.p) <= 0;
}

inline bool operator>=(const char* cs, const MIRGString& s)
{
    return stringCompare(cs, s.p) >= 0;
}

inline int MIRGString::compare(const MIRGString& s) const
{
    return stringCompare(p, s.p);
}

inline int MIRGString::compare(const char* cs) const
{
    return stringCompare(p, cs);
}

inline int MIRGString::compareNoCase(const MIRGString& s) const
{
    return stringICompare(p, s.p);
}

inline int MIRGString::compareNoCase(const char* cs) const
{
    return stringICompare(p, cs);
}

inline int MIRGString::compareSubstring(const MIRGString& s) const
{
    return stringNCompare(p, s.p, s.len);
}

inline int MIRGString::compareSubstring(const char* cs) const
{
    return stringNCompare(p, cs, strlen(cs));
}

inline int MIRGString::compareSubstringNoCase(const MIRGString& s) const
{
    return stringNICompare(p, s.p, s.len);
}

inline int MIRGString::compareSubstringNoCase(const char* cs) const
{
    return stringNICompare(p, cs, strlen(cs));
}

//## MIRGString General methods

inline unsigned MIRGString::length() const
{
    return len;
}

inline bool MIRGString::isEmpty() const
{
    return (len<=0);
}

inline unsigned MIRGString::capacity() const
{
    return (alloc>0?alloc - 1:0);
}

inline void MIRGString::truncate(unsigned new_length)
{
    if ((new_length>=0) && (new_length<len)) {
        len=new_length;
        p[len]='\0';
    }
}


//## MIRGString Conversion

inline MIRGString::operator const char*() const
{
    return p;
}

inline const char* MIRGString::chars() const
{
    return p;
}

inline const unsigned char* MIRGString::unsignedChars() const
{
    return (const unsigned char*)p;
}


//## MIRGString Character manipulation

inline void MIRGString::setAt(unsigned i, const char c)
{
    if (i<len)
        p[i]=c; 
}

inline char MIRGString::getAt(unsigned i) const
{
    return p[(i>len?len:i)];
}


//## MIRGString Character search

inline int MIRGString::find(const char c) const
{
    return find(c, 0);
}
inline int MIRGString::find(const char c, unsigned int pos) const
{
    char* pc = (pos<len)?strchr(p + pos, c):NULL;
    return (pc != NULL) ? pc-p : -1;
}

inline int MIRGString::reverseFind(const char c) const
{
    /*
     * We could use strrchr but since we know the length 
     * of this string and if we actually expect to find this 
     * character it is probably more efficient to start from 
     * the end.
     */

    char* pc = p+len;
    while (pc-->p) {
        if (*pc==c) {
            return pc-p;
        }
    }
    return -1;
}

inline int MIRGString::find(const char* cs) const
{
    return find(cs, 0);
}
inline int MIRGString::find(const char* cs, unsigned int pos) const
{
    char* pstr = (pos<len)?strstr(p + pos, cs):NULL;
    return (pstr != NULL) ? pstr-p : -1;
}





/*
 * MIRString Inline methods
 */

//## MIRString Construction

inline MIRString::MIRString()
{
    len=0;
    alloc=0;
    p=emptyPtr;
}

inline MIRString::MIRString(const char* cs)
{
    init(cs);
}

inline MIRString::MIRString(const unsigned char* cs)
{
    init((const char*)cs);
}

inline MIRString::MIRString(const char* cs, int count)
{
    init(cs, count);
}

inline MIRString::MIRString(const unsigned char* cs, int count)
{
    init((const char*)cs, count);
}

#ifdef WIN32
inline void* MIRString::operator new(size_t size, const char* file, int line)
{
    return operator new(size,_NORMAL_BLOCK,file,line);
}

#if _MSC_VER >= 1200
inline void MIRString::operator delete(void* ptr, int block_type, const char* file, int line)
{
    operator delete(ptr);
}

inline void MIRString::operator delete(void* ptr, const char* file, int line)
{
    operator delete(ptr);
}
#endif
#endif





/*
 * MIRTString Inline methods
 */

//## MIRTString Construction

inline MIRTString::MIRTString()
{
    len=0;
    alloc=0;
    p=emptyPtr;
}

inline MIRTString::MIRTString(const MIRString& s)
{
    //MIRSTRING_STAT(tstr_ct_str++);
    len=s.len;
    alloc=0;
    p=s.p;
}

inline MIRTString::MIRTString(const char* cs)
{
    init(cs);
}

inline MIRTString::MIRTString(const unsigned char* cs)
{
    init((const char*)cs);
}

inline MIRTString::MIRTString(const char* cs, int count)
{
    init(cs, count);
}

inline MIRTString::MIRTString(const unsigned char* cs, int count)
{
    init((const char*)cs, count);
}


#endif /* end of file */
