//####################################################################
//##     Copyright  Meta Integration Technology, Inc. 1997-2002    ##
//##                      All Rights Reserved.                      ##
//####################################################################
//## Meta Integration(tm) Model Bridge (MIMB), Works (MIW),         ##
//## Repository (MIR), Repository SDK (MIRSDK), and Database SDK    ##
//## (MIDBSDK) are trademarks of Meta Integration Technology, Inc.  ##
//## Other product and company names (or logos) mentioned herein    ##
//## may be the trademarks of their respective owners.              ##
//####################################################################
//##                 http://www.metaintegration.com                 ##
//####################################################################

#ifndef MIRVector_h
#define MIRVector_h 1

/*
 * Implements the MIRContainer abstract class using a vector.
 *
 * Notes:
 * - Two versions of the vector are provided. In the first one the 
 *   order in which the objects are inserted in the vector is 
 *   preserved and with the second one it is not.
 */

#include "MIRContainer.h"
#include "MIRArrayInterface.h"

/*
 * MIRVector
 *
 * Notes:
 * - does not preserves the order in which objects are added to the 
 *   vector.
 */
class MIR_STD_API MIRVector: public MIRContainer, public MIRArrayInterface {

  public:

    class Iterator;
    friend class Iterator;

    //## Construction

      MIRVector();
      //    Creates an empty vector.

      MIRVector(int init_size);
      //    Creates a vector that will allocate memory for 
      //    init_size elements when the first element is added.

      MIRVector(const MIRVector& src);
      //    Creates a new vector initialising it with the contents of 
      //    the given source vector. Note that only the object 
      //    pointers are copied, the objects themselves are not duplicated.

      static void* operator new(size_t size);
      static void operator delete(void* ptr);
#ifdef WIN32
      static void* operator new(size_t size, const char* file, int line);
      static void* operator new(size_t size, int block_type, const char* file, int line);
#if _MSC_VER >= 1200
      static void operator delete(void* ptr, const char* file, int line);
      static void operator delete(void* ptr, int block_type, const char* file, int line);
#endif
#endif
      //    Used to allocate/deallocate vectors (and subclasses). 

      virtual MIRVector& operator=(const MIRVector& right);
      //    Replaces the contents of the current vector with that of the
      //    given vector. Note that only the object pointers are copied,
      //    the objects themselves are not duplicated.


    //## Destruction

      virtual ~MIRVector();
      //    Releases the resources used by this vector. Note that this 
      //    operation does not release the memory used by the objects 
      //    that have been put in the vector.


    //## Query methods

      virtual int operator==(const MIRVector& right) const;
      virtual int operator!=(const MIRVector& right) const;
      //    Tests for (in)equality of the contents of two vectors
      //    by comparing the values of the pointers they contain.


    //## General functions

      virtual MIRBoolean empty();
      //    Removes all the elements that were in the vector.

      void shrink();
      //    Shrinks the vector to the minimum size.

      void resize(int array_size);
      //    If the vector already contains elements, the underlying
      //    array is resized (and possibly truncated) to array_size.
      //    Otherwise, the vector will allocate memory for 
      //    array_size elements when the first element is added.
      //    If array_size is < 1, the function does nothing.


    //## Creating and deleting an iterator for this container

      virtual MIRIterator* newIterator(MIRBoolean first=true) const;
      //    Allocates an iterator operating on this container. Note 
      //    that each iterator is usually specific to the container type
      //    that allocated it.
      //    If first is true then the iterator is initialised so that it 
      //    points to the first element in the container. Otherwise it 
      //    designates the last element of the container.
      //    Returns a pointer to the new iterator if successful and
      //    NULL otherwise.

      virtual void deleteIterator(MIRIterator* iterator) const;
      //    Releases an iterator. You must call this method to free any
      //    iterator allocated using the newIterator method. As iterators 
      //    specific to one kind of container be sure to call newIterator
      //    and deleteIterator on the same kind of iterator.


    //## Iterator based methods for adding and removing elements
    //## (Container independant)

      virtual MIRBoolean add(const MIRContainerItem* ptr, MIRBoolean unique=true, MIRIterator* iterator=NULL);
      //    Adds a new object to the vector. If unique is true then
      //    the insertion will fail is an object with the same key
      //    is already in the vector.
      //    If the insertion succeeds and iterator is not NULL then it is
      //    set to the position where the object has been inserted and true
      //    is returned. Otherwise, false is returned and iterator is left
      //    unchanged.

      virtual MIRBoolean contains(const MIRContainerItem* ptr, MIRIterator* iterator=NULL) const;
      //    Returns true if the container contains the object
      //    pointed to by "ptr" and sets the iterator to the position
      //    of this object. Returns false and leaves the iterator 
      //    unchanged otherwise.

      virtual MIRContainerItem* findFirst(const MIRContainerKey* key, MIRIterator* iterator=NULL) const;
      //    Looks for the first object with the given key.
      //    Returns a pointer to this object if found and sets the 
      //    iterator to the position of that object. Repeatedly calling 
      //    this iterator's findNext(key) will allow you to enumerate 
      //    all this container's objects that have the same key.
      //    Returns NULL and leaves the iterator unchanged if no object
      //    with this key was found.

      virtual MIRContainerItem* remove(MIRIterator* iterator);
      //    Removes the element designated by iterator.
      //    Returns a pointer to that object if successful. The iterator 
      //    then points to the element replacing the removed element or 
      //    to the last element in the container if the removed element 
      //    was the last one.
      //    If the operation fails it returns NULL and the iterator is 
      //    left unchanged.
      //
      //    Note:
      //    - This method does not preserve the order in which the 
      //       objects have been added to the vector.


    //## Position based methods for adding and removing elements. 
    //## Implements MIRArrayInterface.

      virtual int addElement(const MIRContainerItem* ptr, MIRBoolean unique=true);
      //    Adds a new object to the container. If unique is true then
      //    the insertion will fail if an object with the same key
      //    is already in the container.
      //    Returns the position of the new element if the insertion 
      //    succeeds and -1 otherwise.

      virtual int containsElement(const MIRContainerItem* ptr) const;
      //    Checks if ptr is already in the container.
      //    If yes then returns its position. 
      //    Returns -1 otherwise.

      virtual MIRContainerItem* operator[](int pos) const;
      //    Returns the object at the given position

      virtual int findFirstElement(const MIRContainerKey* key) const;
      //    Retrieves the first object in the container with the given key. 
      //    Returns the position of this object if successful 
      //    and -1 otherwise. 

      virtual int findLastElement(const MIRContainerKey* key) const;
      //    Retrieves the last object in the container with the given key. 
      //    Returns the position of this object if successful 
      //    and -1 otherwise. 

      virtual int findNextElement(const MIRContainerKey* key, int pos) const;
      //    Looks for an object with the given key at or after
      //    the specified position. Returns its index if successful 
      //    and NULL otherwise.

      virtual int findPreviousElement(const MIRContainerKey* key, int pos) const;
      //    Looks for an object with the given key at or before 
      //    the specified position. Returns its index if successful 
      //    and NULL otherwise.

      virtual MIRContainerItem* removeElement(int pos);
      //    Removes the object contained at position pos. Note that 
      //    this operation may or may not preserve the order in 
      //    which the objects are stored.
      //    Returns a pointer to that object if successful and NULL 
      //    otherwise.

      class MIR_STD_API Iterator: public MIRContainer::Iterator {

      public:

        friend class MIRVector;
        friend class MIROrderedVector;

        // ## Construction

          Iterator();
          //    Creates a new iterator not associated to a container.

          Iterator(const MIRVector& vector, MIRBoolean first=true);
          //    Creates a new iterator pointing to the first element 
          //    in the vector if first is true or the last otherwise.

          Iterator(const MIRVector& vector, int pos);
          //    Creates an iterator corresponding to the given position
          //    of the vector.

          static void* operator new(size_t size);
          static void operator delete(void* ptr);
#ifdef WIN32
          static void* operator new(size_t size, const char* file, int line);
          static void* operator new(size_t size, int block_type, const char* file, int line);
#if _MSC_VER >= 1200
          static void operator delete(void* ptr, const char* file, int line);
          static void operator delete(void* ptr, int block_type, const char* file, int line);
#endif
#endif
          //    Used to allocate/deallocate iterators (and subclasses). 

          
        //## Destruction

          virtual ~Iterator();
          //    Releases the resources used by this iterator.


       //## Query methods

          virtual int operator==(const Iterator& right) const;
          virtual int operator!=(const Iterator& right) const;
          //    Tests for (in)equality between two iterators.


        //## General functions

          virtual Iterator& operator=(const Iterator& right);
          //    Copies the iterator given as a parameter

          virtual MIRContainerItem* operator*() const;
          virtual MIRContainerItem* get() const;
          //    Returns a pointer to the object for the current position
          //    of the iterator or NULL if the iterator is invalid.

          int getPosition() const;
          //    Returns the vector position that this iterator points to.


        //## Methods for moving iterators

          virtual MIRContainerItem* moveTo(MIRVector& vector, MIRBoolean first=true);
          //    Moves the iterator to the specified container.
          //    If first is true then the iterator is initialised so that it 
          //    points to the first element in the container. Otherwise it 
          //    designates the last element of the container.
          //    Returns a pointer to the new iterator if successful and
          //    NULL otherwise.

          virtual MIRContainerItem* moveTo(int pos);
          //    Moves the iterator to the specified position. Returns a
          //    pointer to the object at that position if successful and 
          //    NULL otherwise

          virtual MIRContainerItem* moveToFirst();
          //    Moves the iterator to the first element of the container. 
          //    Returns a pointer to the first object if successful
          //    and NULL otherwise

          virtual MIRContainerItem* moveToLast();
          //    Moves the iterator to the last element of the container. 
          //    Returns a pointer to the last object if successful
          //    and NULL otherwise

          virtual MIRContainerItem* next(int nth=1);
          //    Moves the iterator to the nth element to the "right", starting 
          //    from the current position, and returns the corresponding object.
          //    If this object does not exist, because we reached the last
          //    element, the iterator points to the last element of the 
          //    vector and NULL is returned.

          virtual MIRContainerItem* previous(int nth=1);
          //    Moves the iterator to the nth element to the "left", starting
          //    from the current position, and returns the corresponding object.
          //    If this object does not exist, because we reached the first
          //    element, the iterator points to the first element of the 
          //    vector and NULL is returned.

          virtual MIRContainerItem* findFirst(const MIRContainerKey* key);
          //    Returns the first object with the given key in the vector 
          //    and moves the iterator to the position of that object.
          //    If no such element can be found, NULL is returned and the 
          //    iterator is left unchanged.

          virtual MIRContainerItem* findLast(const MIRContainerKey* key);
          //    Returns the last object with the given key in the vector
          //    and moves the iterator to the position of that object.
          //    If no such element can be found, NULL is returned and the 
          //    iterator is left unchanged.

          virtual MIRContainerItem* findNext(const MIRContainerKey* key);
          //    This method should only be called if you successfully called 
          //    one of the "find" operations on this iterator before.
          //
          //    Returns the next element with the given key starting from the 
          //    current position of the iterator and moves the iterator to the 
          //    position of that object.
          //    If no such element can be found, NULL is returned and the 
          //    iterator points to the last vector element.

          virtual MIRContainerItem* findPrevious(const MIRContainerKey* key);
          //    This method should only be called if you successfully called 
          //    one of the "find" operations on this iterator before.
          //
          //    Returns the previous element with the given key starting from the 
          //    current position of the iterator and moves the iterator to the 
          //    position of that object.
          //    If no such element can be found, NULL is returned and the 
          //    iterator points to the first vector element.


      protected:

          int position;
    };

  protected:

      int array_size;
      MIRContainerItem** array;
};

typedef MIRVector::Iterator MIRVectorIterator;

/*
 * MIROrderedVector
 *
 * Notes:
 * - Preserves the order in which objects are added to the 
 *   vector. As consequence the remove operations are less efficient
 *   than their MIRVector counterparts.
 */
class MIR_STD_API MIROrderedVector: public MIRVector {

  public:

    //## Construction

      MIROrderedVector();
      //    Creates an empty vector.

      MIROrderedVector(int init_size);
      //    Creates a vector that will allocate memory for 
      //    init_size elements when the first element is added.

      MIROrderedVector(const MIROrderedVector &src);
      //    Creates a new vector initialising it with the contents of 
      //    the given source vector. Note that only the object 
      //    pointers are copied, the objects themselves are not duplicated.


    //## Iterator based methods for adding and removing elements
    //## (Container independant)

      virtual MIRBoolean insertElement(const MIRContainerItem* ptr, int pos);
      //    Insert the element at the given position in the vector.
      //    Returns true if successful, false otherwise.

      virtual MIRContainerItem* remove(MIRIterator* iterator);
      //    Removes the element designated by the iterator.
      //    Returns a pointer to that object if successful and NULL 
      //    otherwise.


    //## Position based methods for adding and removing elements
    //## (Specific to vectors)

      MIRContainerItem* remove(int pos);
      //    Removes the object contained at the specified position.
      //    Returns a pointer to that object if successful and NULL 
      //    otherwise.

      typedef MIRVector::Iterator Iterator;
};

typedef MIROrderedVector::Iterator MIROrderedVectorIterator;




/*
 * MIRConstVector
 *
 * Notes:
 * - Wrap a given item array. No modification can be done to
 *   the Vector.
 */
class MIR_STD_API MIRConstVector: public MIRVector {

  public:

    //## Construction

      MIRConstVector(const MIRContainerItem** items, int nb_items);
      //    Creates a new vector initialising it with the contents of 
      //    the given item array. This array must be available as 
      //    long as you are using this vector.


    //## Destruction

      virtual ~MIRConstVector();
      //    Releases the resources used by this vector. Note that this 
      //    operation does not release the memory used by the objects 
      //    that have been put in the vector, or the array that was 
      //    provided for its initialization.

      
    //## General functions

      virtual MIRBoolean empty();
      //    Since this is a const vector this operation is not allowed.

      void resize(int array_size);
      //    Since this is a const vector this operation is not allowed.


    //## Iterator based methods for adding and removing elements
    //## (Container independant)

      virtual MIRBoolean add(const MIRContainerItem* ptr, MIRBoolean unique=true, MIRIterator* iterator=NULL);
      //    Since this is a const vector this operation is not allowed.

      virtual MIRContainerItem* remove(MIRIterator* iterator);
      //    Since this is a const vector this operation is not allowed.


    //## Position based methods for adding and removing elements. 
    //## Implements MIRArrayInterface.

      virtual int addElement(const MIRContainerItem* ptr, MIRBoolean unique=true);
      //    Since this is a const vector this operation is not allowed.

      virtual MIRContainerItem* removeElement(int pos);
      //    Since this is a const vector this operation is not allowed.

      typedef MIRVector::Iterator Iterator;
};

typedef MIRConstVector::Iterator MIRConstVectorIterator;



/*
 * Inline methods
 */

#ifdef WIN32
inline void* MIRVector::operator new(size_t size, const char* file, int line)
{
    return operator new(size,_NORMAL_BLOCK,file,line);
}

#if _MSC_VER >= 1200
inline void MIRVector::operator delete(void* ptr, int block_type, const char* file, int line)
{
    operator delete(ptr);
}

inline void MIRVector::operator delete(void* ptr, const char* file, int line)
{
    operator delete(ptr);
}
#endif
#endif


inline MIRIterator* MIRVector::newIterator(MIRBoolean first) const
{
    return _new_dbg MIRVector::Iterator(*this,first);
}

inline void MIRVector::deleteIterator(MIRIterator* iterator) const
{
    delete iterator;
}

#ifdef WIN32
inline void* MIRVector::Iterator::operator new(size_t size, const char* file, int line)
{
    return operator new(size,_NORMAL_BLOCK,file,line);
}

#if _MSC_VER >= 1200
inline void MIRVector::Iterator::operator delete(void* ptr, int block_type, const char* file, int line)
{
    operator delete(ptr);
}

inline void MIRVector::Iterator::operator delete(void* ptr, const char* file, int line)
{
    operator delete(ptr);
}
#endif
#endif

inline MIRVector::Iterator::~Iterator()
{
}

inline MIRBoolean MIRConstVector::empty()
{
    return false;
}

inline void MIRConstVector::resize(int new_size)
{
}

inline MIRBoolean MIRConstVector::add(const MIRContainerItem* ptr, MIRBoolean unique, MIRIterator* iterator)
{
    return false;
}

inline MIRContainerItem* MIRConstVector::remove(MIRIterator* iterator)
{
    return NULL;
}

inline int MIRConstVector::addElement(const MIRContainerItem* ptr, MIRBoolean unique)
{
    return -1;
}

inline MIRContainerItem* MIRConstVector::removeElement(int pos)
{
    return NULL;
}

#endif /* end of file */
